"""BioAT package.
BioAT can be a package to import.
BioAT also can be a command-line tool.
It is a bioinformatic tool/pkg bundle for python.
"""
# !use "# ruff: isort: skip_file" annotation to skip this file when sorting imports
# ruff: isort: skip_file
from bioat._meta import (
    __PKG_NAME__,
    __AUTHOR__,
    __AUTHOR_EMAIL__,
    __DESCRIPTION__,
    __DOC_FORMAT__,
    __DOC_PAGE__,
    __HOME_PAGE__,
    __ISSUE_PAGE__,
    __LICENSE__,
    __VERSION__,
)

from bioat.bamtools import BamTools
from bioat.bedtools import BedTools
from bioat.crisprtools import CrisprTools

from bioat.fastxtools import FastxTools
from bioat.foldtools import FoldTools
from bioat.hictools import HiCTools

from bioat.metatools import MetaTools
from bioat.searchtools import SearchTools
from bioat.systemtools import SystemTools
from bioat.tabletools import TableTools
from bioat.target_seq import TargetSeq


__all__ = [
    "__PKG_NAME__",
    "__AUTHOR__",
    "__AUTHOR_EMAIL__",
    "__DESCRIPTION__",
    "__DOC_FORMAT__",
    "__DOC_PAGE__",
    "__HOME_PAGE__",
    "__ISSUE_PAGE__",
    "__LICENSE__",
    "__VERSION__",
    "BamTools",
    "BedTools",
    "CrisprTools",
    "FastxTools",
    "FoldTools",
    "HiCTools",
    "MetaTools",
    "SearchTools",
    "SystemTools",
    "TableTools",
    "TargetSeq",
]



"""module of bamtools

bioat bamtools <command> deal with SAM or BAM files

example 1:
    bioat list
        <in shell>:
            $ bioat bam remove_clip --help
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> help(bioat.bam.remove_clip)

example 2:
    _example_
"""

import gzip
import os
import random
import string
import sys
from io import TextIOWrapper
from multiprocessing import Process
from signal import SIG_DFL, SIGPIPE, signal
from sys import stdin as STDIN
from sys import stdout as STDOUT

from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.bamtools")

def setup_signal_handling():
    """Setup signal handlers."""
    signal(SIGPIPE, SIG_DFL)

setup_signal_handling()

class BamTools:
    """Bam toolbox."""

    lm.set_names(cls_name="BamTools")
    def __init__(self):
        pass

    def mpileup2table(
        self,
        mpileup: str,
        output: str | TextIOWrapper = STDOUT,
        threads: int = os.cpu_count(),
        mutation_number_threshold: int = 0,
        temp_dir: str = "__bioat_temp_dir",
        remove_temp: bool = True,
        log_level: str = "WARNING",
    ) -> None:
        """Converts an mpileup file to a structured info file.

        Args:
            mpileup (str):
                Path to the samtools mpileup format file.
            output (str | TextIOWrapper):
                Path to the output file where parsed data will be stored. Defaults to standard output.
            threads (int):
                Number of threads to utilize for processing. Default is one less than the number of available CPU cores.
            mutation_number_threshold (int):
                Threshold for mutation information; set to 0 to include all sites.
            temp_dir (str, optional):
                Directory for temporary files. Defaults to a directory in '__bioat_temp_dir'.
            remove_temp (bool, optional):
                Indicator for whether to remove temporary files after processing. Defaults to True.
            log_level (str, optional):
                Level of logging. Options are 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'.
                Defaults to 'WARNING'.

        Returns:
            None: This function does not return a value. It outputs a file based on the provided parameters.
        """
        lm.set_names(func_name="mpileup2table")
        lm.set_level(log_level)

        def _temp_split_bam(mpileup, threads, temp_dir):
            # creat temp file name and open file
            lm.logger.debug("Making temp files...")

            input_file_basename = os.path.basename(mpileup)

            temp_filename_list = []
            temp_file_list = []

            os.makedirs(temp_dir, exist_ok=True)

            for index in range(threads):
                # make filename
                temp_file_basename = (
                    "temp_"
                    + input_file_basename
                    + "."
                    + str(index)
                    + "."
                    + "".join(random.sample(string.ascii_letters + string.digits, 16))
                    + ".gz"
                )
                temp_file_name = os.path.join(temp_dir, temp_file_basename)
                temp_filename_list.append(temp_file_name)

                # open file
                temp_file = gzip.open(temp_file_name, "wt")
                temp_file_list.append(temp_file)

            # counting input file line number
            lm.logger.debug("Counting input file...")

            total_input_line_num = 0
            input_file = (
                gzip.open(mpileup, "rt")
                if mpileup.endswith(".gz")
                else open(mpileup, "rt")
            )

            # count total_input_line_num
            for _ in input_file:
                total_input_line_num += 1

            if total_input_line_num % threads == 0:
                each_file_line_num = total_input_line_num // threads
            else:
                each_file_line_num = (total_input_line_num // threads) + 1

            input_file.close()
            lm.logger.debug("Done!")

            # split temp files
            # write into output filename
            input_file = (
                gzip.open(mpileup, "rt")
                if mpileup.endswith(".gz")
                else open(mpileup, "rt")
            )

            for index, line in enumerate(input_file):
                file_index = index // each_file_line_num
                temp_file_list[file_index].write(line)

            # close output filename
            input_file.close()
            [temp_file.close() for temp_file in temp_file_list]
            lm.logger.debug("Make temp files done!")

            return temp_filename_list

        def _merge_out_bmat(temp_filename_list, output, remove_temp=True):
            if type(output) is str:
                output = (
                    gzip.open(output, "wt")
                    if output.endswith(".gz")
                    else open(output, "wt")
                )
            # write header
            header = [
                "chr_name",
                "chr_index",
                "ref_base",
                "A",
                "G",
                "C",
                "T",
                "del_count",
                "insert_count",
                "ambiguous_count",
                "deletion",
                "insertion",
                "ambiguous",
                "mut_num",
            ]
            output.write("\t".join(header) + "\n")

            # open temp bmat file
            for temp_filename in temp_filename_list:
                # write file
                with gzip.open(temp_filename + ".bmat.gz", "rt") as temp_file:
                    for line in temp_file:
                        output.write(line)

            # remove temp files
            if remove_temp:
                for temp_filename in temp_filename_list:
                    os.remove(temp_filename)
                    os.remove(temp_filename + ".bmat.gz")
                os.rmdir(temp_dir)

        # make temp files
        temp_filename_list = _temp_split_bam(
            mpileup=mpileup, threads=threads, temp_dir=temp_dir
        )
        # multiple processing part
        lm.logger.debug("Parsing files...")

        procs_list = []

        for index, temp_mpileup in enumerate(temp_filename_list):
            temp_output = temp_mpileup + ".bmat.gz"
            # add multiprocess
            sub_proc = Process(
                target=_run_mpileup_to_table,
                args=(
                    temp_mpileup,
                    temp_output,
                    mutation_number_threshold,
                ),
            )
            procs_list.append(sub_proc)
            sub_proc.start()

        for sub_proc in procs_list:
            sub_proc.join()
        lm.logger.debug("Done!")
        # merge output files
        lm.logger.debug("Merging files...")
        _merge_out_bmat(
            temp_filename_list=temp_filename_list,
            output=output,
            remove_temp=remove_temp,
        )
        lm.logger.debug("Done!...")

    def remove_clip(
        self,
        input: str | TextIOWrapper = STDIN,
        output: str | TextIOWrapper = STDOUT,
        threads: int = os.cpu_count(),
        output_fmt: str = "SAM",
        remove_as_paired: bool = False,
        max_clip: int = 0,
        log_level: str = "WARNING",
    ):
        """Remove soft/hard clipped reads from a BAM/SAM file.

        This method removes soft/hard clipped reads from a BAM/SAM file. It can
        accept input from stdin and produce output to stdout.

        Args:
            input (str | TextIOWrapper): 
                BAM file sorted by query name with soft/hard clipped reads. 
                Pipe stdin is supported, e.g.:
                [samtools view -h foo_sort_name.bam | bioat bam remove_clip <flags>].
            output (str | TextIOWrapper): 
                BAM file sorted by query name without soft/hard clipped reads. 
                Pipe stdout is supported, e.g.:
                [bioat bam remove_clip <flags> | wc -l] 
                or [bioat bam remove_clip <flags> | samtools view ....].
            threads (int, optional): 
                Number of threads used by pysam and samtools core. 
                Defaults to the number of CPU cores.
            output_fmt (str, optional): 
                Format of the output file, can be "BAM" or "SAM". Defaults to "SAM".
            remove_as_paired (bool, optional): 
                Flag to determine whether to remove single clipped reads. 
                If True, removes both the clipped read and its paired read. The input 
                BAM/SAM must be sorted by name and have header [SO:queryname]. 
                If False, only removes the single clipped read.
            max_clip (int, optional): 
                The maximum number of clips allowed per read. Defaults to 0.
            log_level (str, optional): 
                Logging level for the process. Can be one of 
                'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'. 
                Defaults to "INFO".

        Returns:
            None
        """
        lm.set_names(func_name="remove_clip")
        lm.set_level(log_level)

        try:
            import pysam
        except ImportError:
            lm.logger.error(
                'remove_clip requires "pysam" as the backend for parsing BAM files. Please install it first.'
            )
            sys.exit(0)
        save = pysam.set_verbosity(
            0
        )  # https://github.com/pysam-developers/pysam/issues/939

        if isinstance(input, TextIOWrapper):
            bam_in = pysam.AlignmentFile("-", "rb", check_sq=False)
        else:
            bam_in = pysam.AlignmentFile(input, "rb", check_sq=False)
        pysam.set_verbosity(save)

        try:
            so = bam_in.header["HD"]["SO"]
        except KeyError:
            so = None

        if so:
            if so != "queryname":
                if remove_as_paired:
                    lm.logger.error(
                        "When remove_as_paired is True, the input BAM|SAM must be sorted by"
                        "name and has header [SO:queryname]!\n"
                        f"your header: [SO:{so}]\n"
                    )
                    exit(1)
                else:
                    if so != "coordinate":
                        lm.logger.warning(
                            "the input BAM|SAM must have header [SO:coordinate] or [SO:queryname]!\n"
                            f"your header: [SO:{so}]\n"
                        )

        else:
            lm.logger.warning(
                "the input BAM|SAM must be sorted by name and has header [SO:queryname]!\n"
                "your bam file does not have a header\n"
            )
            if isinstance(input, TextIOWrapper):
                lm.logger.warning(
                    "you can add header to your input file by using `samtools view -h input.bam > input_with_header.bam`"
                )

        write_mode = "wb" if output_fmt.upper() == "BAM" else "w"
        bam_out = pysam.AlignmentFile(output, write_mode, template=bam_in)

        # Iterate through reads.
        if remove_as_paired:
            lm.logger.info(
                "remove_as_paired is True, all paired clipped reads will be removed."
            )
            read1, read2 = None, None

            for read in bam_in:
                if not read.is_paired or read.is_unmapped or read.mate_is_unmapped:
                    # or read.is_duplicate:
                    # only use mapped paired reads
                    continue

                # ------------------------------------------------------------------->>>>>>>>>>
                # 如果先看到read1，将read1赋给read， 判断是否read1、read2都存在
                #    不都存在：
                #        即read2不存在，进入下个循环
                #    都存在，判断query name是否一致
                # ------------------------------------------------------------------->>>>>>>>>>
                if read.is_read2:
                    # if first is read2, then use it to find read2
                    read2 = read
                else:
                    # if first is read1, then use it to find read1, and next query
                    read1 = read
                    read2 = None
                    continue

                # if read1 not None and read2 not None pass to below
                if read1 and read2 and read1.query_name == read2.query_name:
                    # print("found a pair!")
                    # print(f'Cigar: Read1-{read1.cigarstring}, Read2-{read1.cigarstring}')
                    # print(read.get_cigar_stats())
                    softclip1 = read1.get_cigar_stats()[0][4]
                    softclip2 = read2.get_cigar_stats()[0][4]

                    if not remove_as_paired:
                        if softclip1 <= max_clip:
                            bam_out.write(read1)
                        if softclip2 <= max_clip:
                            bam_out.write(read2)
                    else:
                        if softclip1 <= max_clip and softclip2 <= max_clip:
                            bam_out.write(read1)
                            bam_out.write(read2)
                else:
                    lm.logger.fatal(
                        "something wrong with read1/2 pairs, they may have different read id or there is None in them"
                    )
                    lm.logger.fatal(f"read1={read1}, read2={read2}")
                    lm.logger.fatal(
                        f"read1_id={read1.query_name}, read2_id={read2.query_name}"
                    )
                    exit(1)
        else:
            for read in bam_in:
                if read.is_unmapped:
                    # or read.is_duplicate:
                    # only use mapped reads
                    continue

                softclip = read.get_cigar_stats()[0][4]

                if softclip <= max_clip:
                    bam_out.write(read)
        bam_in.close()
        bam_out.close()


def _run_mpileup_to_table(temp_mpileup, temp_output, mutation_number_threshold):
    # open input file
    temp_mpileup = (
        gzip.open(temp_mpileup, "rt")
        if temp_mpileup.endswith(".gz")
        else open(temp_mpileup, "rt")
    )
    # open output file
    temp_output = (
        gzip.open(temp_output, "wt")
        if temp_output.endswith(".gz")
        else open(temp_output, "wt")
    )

    # parse mpileup file
    for line in temp_mpileup:
        data = line.strip().split("\t")
        chr_name = data[0]
        bp = data[1]
        bases = data[4].upper()
        ref = data[2].upper()
        types = {"A": 0, "G": 0, "C": 0, "T": 0, "-": [], "+": [], "Not": []}

        # coverage > 0
        i = 0
        while i < len(bases):
            base = bases[i]

            if base == "^":
                i += 2

            elif base == "$":
                i += 1

            elif base == "-":
                i += 1
                del_str_num = ""

                while bases[i].isdigit():
                    del_str_num += bases[i]
                    i += 1

                delNum = int(del_str_num)
                delSeq = ""
                for a in range(delNum):
                    delSeq += bases[i]
                    i += 1
                types["-"].append(delSeq)

            elif base == "*":
                types["-"].append(bases[i])
                i += 1

            elif base == "+":
                i += 1
                add_str_num = ""

                while bases[i].isdigit():
                    add_str_num += bases[i]
                    i += 1

                addNum = int(add_str_num)
                addSeq = ""
                for a in range(addNum):
                    addSeq += bases[i]
                    i += 1
                types["+"].append(addSeq)

            elif (base == ".") or (base == ","):
                types[ref] += 1
                i += 1

            else:
                if base in types:
                    types[base] += 1
                else:
                    types["Not"].append(base)
                i += 1

        adds = "."
        adds_count = 0
        if len(types["+"]) > 0:
            adds = ",".join(types["+"])
            adds_count = len(types["+"])

        dels = "."
        dels_count = 0
        if len(types["-"]) > 0:
            dels = ",".join(types["-"])
            dels_count = len(types["-"])

        amb = "."
        amb_count = 0
        if len(types["Not"]) > 0:
            amb = ",".join(types["Not"])
            amb_count = len(types["Not"])

        # get other mutation number
        if ref in types:
            mut_num = types["A"] + types["T"] + types["G"] + types["C"] - types[ref]
        else:
            mut_num = 0

        out_list = [
            chr_name,
            bp,
            ref,
            types["A"],
            types["G"],
            types["C"],
            types["T"],
            dels_count,
            adds_count,
            amb_count,
            dels,
            adds,
            amb,
            mut_num,
        ]

        # make a filter
        if mutation_number_threshold:
            if mut_num >= mutation_number_threshold:
                temp_output.write("\t".join(map(str, out_list)) + "\n")
        else:
            temp_output.write("\t".join(map(str, out_list)) + "\n")

    temp_mpileup.close()
    temp_output.close()


if __name__ == "__main__":
    pass



"""Library for alignment, depends on Biopython

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

Library for alignment, depends on Biopython

example 1:
    bioat.lib.libalignment
        <in python consolo>:
            >>> from bioat.lib.libalignment import instantiate_pairwise_aligner
            >>> aligner = instantiate_pairwise_aligner()
            >>> alignment = get_best_alignment(
            >>>     Seq("GGCACTGCGGCTGGAAAAAAAAAAAAAAAGT"),
            >>>     Seq("GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT"),
            >>>     aligner=aligner,
            >>>     consider_strand=True
            >>> )
            >>> res = get_aligned_seq(alignment, reverse=False)
            >>> print(res)
            >>> res = get_aligned_seq(alignment, reverse=True)
            >>> print(res)
            >>> res = get_alignment_info(alignment, reverse=False)
            >>> print(res)
            >>> res = get_alignment_info(alignment, reverse=True)
            >>> print(res)
"""


from Bio.Align import Alignment, PairwiseAligner, substitution_matrices
from Bio.Seq import Seq

from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.lib.libalignment")


def instantiate_pairwise_aligner(
    scoring_match: float = 1,
    penalty_mismatch: float = 0.8,
    penalty_gap_open: float = -5,
    penalty_gap_extension: float = -2,
    penalty_query_left_gap_score: float = 0,
    penalty_query_right_gap_score: float = 0,
    mode="global",
    letn_match=False,
    score_matrix_dict: None | dict = None,
    log_level="DEBUG",
) -> PairwiseAligner:
    """Returns a PairwiseAligner object.

    Args:
        scoring_match (float, optional): Scoring for matches. Defaults to 1.
        penalty_mismatch (float, optional): Penalty for mismatches. Defaults to 0.8.
        penalty_gap_open (float, optional): Penalty for opening a gap. Defaults to -5.
        penalty_gap_extension (float, optional): Penalty for extending a gap. Defaults to -2.
        penalty_query_left_gap_score (float, optional): Penalty for left query gap. Defaults to 0.
        penalty_query_right_gap_score (float, optional): Penalty for right query gap. Defaults to 0.
        mode (str, optional): Alignment mode, either "global" or "local". Defaults to "global".
        letn_match (bool, optional): Whether to treat base 'N' as a match. Defaults to False.
        log_level (str, optional): Logging level. Defaults to "DEBUG".
        score_matrix_dict (dict, optional): Custom scoring matrix for bases or amino acids. Defaults to None.

    Returns:
        PairwiseAligner: An object instantiated from PairwiseAligner.

    Details for `score_matrix_dict`:
        You can use any base and any score, such as AGCTN, or 20AA + 2 abnormal AAs, 
        or any characters you want. For example:

            score_matrix_dict = {

                ("A", "A"): scoring_match,
                ("A", "G"): penalty_mismatch,
                ("A", "C"): penalty_mismatch,
                ("A", "T"): penalty_mismatch,
                ("A", "N"): scoring_match,
                ("G", "A"): penalty_mismatch,
                ("G", "G"): scoring_match,
                ("G", "C"): penalty_mismatch,
                ("G", "T"): penalty_mismatch,
                ("G", "N"): scoring_match,
                ("C", "A"): penalty_mismatch,
                ("C", "G"): penalty_mismatch,
                ("C", "C"): scoring_match,
                ("C", "T"): penalty_mismatch,
                ("C", "N"): scoring_match,
                ("T", "A"): penalty_mismatch,
                ("T", "G"): penalty_mismatch,
                ("T", "C"): penalty_mismatch,
                ("T", "T"): scoring_match,
                ("T", "N"): scoring_match,
                ("N", "A"): scoring_match,
                ("N", "G"): scoring_match,
                ("N", "C"): scoring_match,
                ("N", "T"): scoring_match,
                ("N", "N"): scoring_match

            }
    """
    lm.set_names(func_name="instantiate_pairwise_aligner")
    lm.set_level(log_level)

    aligner = PairwiseAligner()

    if not letn_match and score_matrix_dict is None:
        # !如果 letn_match为False，
        # !并且未指定 score_matrix_dict
        # !则考虑本流程
        # !否则进入else中的自定义流程
        aligner.mode = mode
        aligner.match = scoring_match
        aligner.mismatch = penalty_mismatch
        aligner.open_gap_score = penalty_gap_open
        aligner.extend_gap_score = penalty_gap_extension
        aligner.query_left_gap_score = penalty_query_left_gap_score
        aligner.query_right_gap_score = penalty_query_right_gap_score
        aligner.wildcard = None
    else:
        if not score_matrix_dict:
            score_matrix_dict = {
                ("A", "A"): scoring_match,
                ("A", "G"): penalty_mismatch,
                ("A", "C"): penalty_mismatch,
                ("A", "T"): penalty_mismatch,
                ("A", "N"): scoring_match,
                ("G", "A"): penalty_mismatch,
                ("G", "G"): scoring_match,
                ("G", "C"): penalty_mismatch,
                ("G", "T"): penalty_mismatch,
                ("G", "N"): scoring_match,
                ("C", "A"): penalty_mismatch,
                ("C", "G"): penalty_mismatch,
                ("C", "C"): scoring_match,
                ("C", "T"): penalty_mismatch,
                ("C", "N"): scoring_match,
                ("T", "A"): penalty_mismatch,
                ("T", "G"): penalty_mismatch,
                ("T", "C"): penalty_mismatch,
                ("T", "T"): scoring_match,
                ("T", "N"): scoring_match,
                ("N", "A"): scoring_match,
                ("N", "G"): scoring_match,
                ("N", "C"): scoring_match,
                ("N", "T"): scoring_match,
                ("N", "N"): scoring_match,
            }

        substitution_matrix = substitution_matrices.Array(
            data=score_matrix_dict
        )
        aligner.mode = mode
        aligner.substitution_matrix = substitution_matrix
        # !自定义矩阵时，不可定义match和mismatch的打分，不然会出问题，在矩阵中给分即可
        # !aligner.match = scoring_match
        # !aligner.mismatch = penalty_mismatch
        aligner.open_gap_score = penalty_gap_open
        aligner.extend_gap_score = penalty_gap_extension
        aligner.query_left_gap_score = penalty_query_left_gap_score
        aligner.query_right_gap_score = penalty_query_right_gap_score
        aligner.wildcard = None
        # !自定义的打分矩阵中如果有"N"，则wildcard最好注释掉
        # !wildcard是一个特殊字符，用于在序列比对过程中表示一个可以与任何字符匹配的通配符
        # !设置wildcard的目的通常是为了在处理包含未知或可变字符的序列时提高比对的灵活性
        # !使用wildcard时需要谨慎，因为它可能会影响比对的准确性和生物学意义
        # !可以将wildcard设置为'N'，这样在比对过程中，字符'N'将被视为与任何字符都匹配

    lm.logger.info(aligner)
    return aligner


def get_best_alignment(
    seq_a: Seq, seq_b: Seq, aligner: PairwiseAligner, consider_strand=True
):
    """Get best alignment by heightest alignment score

    Get best alignment by heightest alignment score,
    default to consider strand, that means, it will attempt to use
    seq_b.reverse_complement() to align

    :param seq_a: seq_a
    :type seq_a: Seq
    :param seq_b: seq_b
    :type seq_b: Seq
    :param aligner: aligner returns from instantiate_pairwise_aligner
    :type aligner: PairwiseAligner
    :param consider_strand: consider to use seq_b.reverse_complement() or not, defaults to True
    :type consider_strand: bool, optional
    :return: Alignment with a new attribute: alignment.is_a_reverse_complement_alignment (True / False)
    :rtype: Alignment
    """
    alignments = aligner.align(seq_a, seq_b)
    alignment = sorted(alignments, key=lambda x: x.score, reverse=True)[0]
    is_a_reverse_complement_alignment = None
    final_alignment = None

    if consider_strand:
        alignments_rc = aligner.align(seq_a, seq_b.reverse_complement())
        alignment_rc = sorted(
            alignments_rc, key=lambda x: x.score, reverse=True
        )[0]

        if alignment.score >= alignment_rc.score:
            final_alignment = alignment
            is_a_reverse_complement_alignment = False
        else:
            final_alignment = alignment_rc
            is_a_reverse_complement_alignment = True
    else:
        final_alignment = alignment
        is_a_reverse_complement_alignment = False

    # obj add attr
    final_alignment.is_a_reverse_complement_alignment = (
        is_a_reverse_complement_alignment
    )
    return final_alignment


def get_aligned_seq(
    alignment: Alignment, reverse: bool = False, letn_match=False
) -> dict:
    """Retrieve aligned sequence information from a Bio.Align.Alignment object.

    Args:
        alignment (Bio.Align.Alignment): 
            A `Bio.Align.Alignment` object containing the alignment information.
        reverse (bool, optional): 
            Whether to return reversed sequence information. Defaults to False.

    Returns:
        dict: A dictionary containing the alignment details with the following keys:
            - 'reference_seq' (str): The aligned reference sequence.
            - 'aln_info' (str): The alignment information string, indicating matches ('|'), 
              mismatches ('.'), and gaps ('-').
            - 'target_seq' (str): The aligned target sequence.

    Example:
        >>> result = get_aligned_seq(alignment)
        >>> print(result)
        {
            'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT',
            'aln_info':      '--..|.|||||||||||||||||||||||--||',
            'target_seq':    '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'
        }

    """
    ls = []

    for x in str(alignment).strip().split("\n"):
        if x:
            ls.append(x[20:].split(" ")[0])
    if not reverse:
        res = dict(
            reference_seq="".join(ls[::3]),
            aln_info="".join(ls[1::3]),
            target_seq="".join(ls[2::3]),
        )
    else:
        res = dict(
            reference_seq="".join(ls[::3])[::-1],
            aln_info="".join(ls[1::3])[::-1],
            target_seq="".join(ls[2::3])[::-1],
        )
    if letn_match:
        length = len(res["reference_seq"])
        # change base N to match
        i_n_a = [
            i
            for i in range(length)
            if (res["reference_seq"][i] == "N")
            and (res["target_seq"][i] != "-")
        ]
        i_n_b = [
            i
            for i in range(length)
            if (res["target_seq"][i] == "N")
            and (res["reference_seq"][i] != "-")
        ]
        i_n = list(set(i_n_a) | set(i_n_b))
        aln_list = list(res["aln_info"])

        for i in sorted(i_n):
            aln_list[i] = "★"
        res["aln_info"] = "".join(aln_list)

    return res


def get_alignment_info(
    alignment: Alignment,
    reverse: bool = False,
    letn_match=False,
    log_level="WARNING",
):
    """Analyze a Bio.Align.Alignment object to extract alignment metrics.

    This function calculates the match count, mismatch count, gap count, alignment score,
    and start and end indices for a given alignment.

    Args:
        alignment (Bio.Align.Alignment): A `Bio.Align.Alignment` object containing the alignment information.
        reverse (bool, optional): Whether to return reversed sequence information. Defaults to False.

    Returns:
        list: A list containing the following metrics:

            - match_count (int): The number of matches in the alignment.
            - mismatch_count (int): The number of mismatches in the alignment.
            - gap_count (int): The number of gaps within the sgRNA alignment region. Note: This counts only gaps within the alignment region, not total gaps.
            - alignment_score (float): The alignment score.
            - start_index (int): The start index of the sgRNA alignment region.
            - end_index (int): The end index of the sgRNA alignment region.

    Notes:
        The gap count specifically counts gaps within the sgRNA alignment region, not the total gaps in the entire sequence. 
        For example:

            >>> Reference:  AGTGGTAAGAAGAAGACGAGACATAATGAG
            >>>             ------||||||||||||||.|----||||
            >>> Target:     ------AAGAAGAAGACGAGCC----TGAG

        In this example: Gap count = 4 (within sgRNA alignment region), not 10 (total gaps). 


    Example:
        The returned list will be as follows: return_list = [match_count, mismatch_count, gap_count, alignment_score, start_index, end_index]

    """
    lm.set_names(func_name="get_alignment_info")
    lm.set_level(log_level)

    if reverse:
        aln_res = get_aligned_seq(
            alignment, reverse=True, letn_match=letn_match
        )
    else:
        aln_res = get_aligned_seq(alignment, letn_match=letn_match)

    reference_seq = aln_res["reference_seq"]
    aln_info = aln_res["aln_info"]
    target_seq = aln_res["target_seq"]
    aln_score = alignment.score
    target_seq_length = len(target_seq)
    target_seq_length_rm_gap = len(target_seq.replace("-", ""))
    aligned_coordinates = alignment.aligned[::-1]
    if hasattr(alignment, "is_a_reverse_complement_alignment"):
        is_a_reverse_complement_alignment = (
            alignment.is_a_reverse_complement_alignment
        )
    else:
        is_a_reverse_complement_alignment = None

    lm.logger.debug(f"reference_seq = {reference_seq}")
    lm.logger.debug(f"aln_info      = {aln_info}")
    lm.logger.debug(f"target_seq    = {target_seq}")
    lm.logger.debug(f"aln_score     = {aln_score}")
    lm.logger.debug(f"target_seq_length (remove gap) = {target_seq_length_rm_gap}")
    lm.logger.debug(
        f"is_a_reverse_complement_alignment = {is_a_reverse_complement_alignment}"
    )
    lm.logger.debug(f"aligned_coordinates =\n{aligned_coordinates}")

    # define params
    ref_aln_start = target_seq_length - len(target_seq.lstrip("-"))
    ref_aln_end = ref_aln_start + len((target_seq.strip("-")))
    match_count = 0
    mismatch_count = 0
    gap_count = 0
    parsed_target_bases = 0
    encounter_target_seq = False

    for idx, base_info in enumerate(aln_info):
        """
        ---TTCTGCCTCTGGAGAGGG-GGAGGGGCCT---
        -------|.|.|||..|-.||-||.||--------
        -------GGCACTGCGG-TGGAGGTGG--------
        """
        if not encounter_target_seq:
            if target_seq[idx] == "-":
                # if start with -
                # not encounter with target_seq yet!
                # print(idx, base_info)
                continue
            else:
                # find first aligned base
                # encounter with target_seq!
                encounter_target_seq = True

        if encounter_target_seq and (
            parsed_target_bases < target_seq_length_rm_gap
        ):
            # from the same loop of [else: catch_start = True]
            # parse start base and later base
            # print(parsed_target_bases, target_seq_length_rm_gap, base_info)
            if base_info == "|" or base_info == "★":
                # match
                match_count += 1
                parsed_target_bases += 1
            elif base_info == ".":  # . instead of X after Biopython 1.8.0
                # mismatch
                mismatch_count += 1
                parsed_target_bases += 1
            elif base_info == "-":
                # gap on reference_seq or target_seq
                gap_count += 1

                if reference_seq[idx] == "-":
                    # gap on reference_seq
                    # value base on target_seq
                    parsed_target_bases += 1
                    pass
                else:
                    # gap on target_seq
                    # no value base on target_seq
                    parsed_target_bases += 0
                    pass
            else:
                lm.logger.critical(
                    f"find not surpported alignment symbol: {base_info} ({idx} on the reference), check"
                    f"whether Biopython >= 1.8.0 or not"
                )
                exit(1)
        else:
            # parsed_target_bases = target_seq_length_rm_gap
            # full target_seq bases were processed
            break

    res = dict(
        match_count=match_count,
        mismatch_count=mismatch_count,
        gap_count=gap_count,
        aln_score=aln_score,
        ref_aln_start=ref_aln_start,
        ref_aln_end=ref_aln_end,
        is_a_reverse_complement_alignment=is_a_reverse_complement_alignment,
        alignment=aln_res,
    )
    return res



import logging
import logging.handlers
import os

import coloredlogs

from .exceptions import BioatInvalidParameterError

__all__ = ["LoggerManager"]


class LoggerManager:
    LOG_FORMAT = "%(asctime)s.%(msecs)03d - [%(name)s] - %(filename)s[line:%(lineno)4d] - %(levelname)+8s: %(message)s"

    def __init__(
        self,
        log_level: str = "ERROR",
        mod_name: str = "bioat.logger",
        cls_name: str | None = None,
        func_name: str | None = None,
    ):
        """
        Initialize the LoggerManager with default log level and module name.

        Args:
            log_level (str, optional): Default logger level. Defaults to "ERROR".
            mod_name (str, optional): Module name. Defaults to "fbt".
            cls_name (str, optional): The name of the class for which the logger is created.
                Defaults to None.
            func_name (str, optional): The name of the function for which the logger is created.
                Defaults to None.
        """
        self.log_level = self._get_log_level(log_level)
        self.mod_name = mod_name
        self.cls_name = cls_name
        self.func_name = func_name
        self.logger = self._create_logger()
        self.logger.propagate = False

    def _get_log_level(self, log_level: str) -> int:
        """Parse and return the logging level."""
        if log_level.upper() not in logging._nameToLevel:
            raise BioatInvalidParameterError(
                f"Invalid log level: {log_level}. "
                "Choose from CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET."
            )
        return logging._nameToLevel[log_level.upper()]

    def _get_logger_name(self) -> str:
        """Construct the logger name based on module, class, and function."""
        if self.cls_name and self.func_name:
            return f"{self.mod_name}.{self.cls_name}.{self.func_name}"
        elif self.cls_name:
            return f"{self.mod_name}.{self.cls_name}"
        elif self.func_name:
            return f"{self.mod_name}.{self.func_name}"
        else:
            return self.mod_name

    def _create_logger(self) -> logging.Logger:
        """Create and configure the logger."""
        name = self._get_logger_name()
        logger = logging.getLogger(name)

        # Console handler with coloredlogs
        coloredlogs.DEFAULT_FIELD_STYLES = dict(
            asctime=dict(color="yellow", bold=True),
            name=dict(color="blue", bold=True),
            filename=dict(color="cyan"),
            lineno=dict(color="green", bold=True),
            levelname=dict(color="magenta", bold=True),
        )
        coloredlogs.DEFAULT_LEVEL_STYLES = dict(
            critical=dict(color="red", bold=True),
            error=dict(color="red"),
            warning=dict(color="yellow"),
            info=dict(color="green"),
            debug=dict(color="blue"),
        )
        coloredlogs.install(
            fmt=self.LOG_FORMAT,  # 使用你的自定义日志格式
            level=self.log_level,
            logger=logger,
            field_styles=dict(
                asctime=dict(color="yellow", bold=True),
                name=dict(color="blue", bold=True),
                filename=dict(color="cyan"),
                lineno=dict(color="green", bold=True),
                levelname=dict(color="magenta", bold=True),
            ),
            level_styles=dict(
                critical=dict(color="red", bold=True),
                error=dict(color="red"),
                warning=dict(color="yellow"),
                info=dict(color="green"),
                debug=dict(color="blue"),
            ),
        )
        logger.propagate = False
        return logger

    def set_names(self, cls_name: str | None = None, func_name: str | None = None):
        """
        Update the class and/or function context for the logger.

        Args:
            cls_name (str, optional): New class name.
            func_name (str, optional): New function name.
        """
        self.cls_name = cls_name
        self.func_name = func_name
        self.logger.name = self._get_logger_name()
        self.logger.propagate = False

    def set_level(self, log_level: str):
        """
        Update the logging level dynamically.

        Args:
            log_level (str): The new logging level.
        """
        self.log_level = self._get_log_level(log_level)
        self.logger.setLevel(self.log_level)
        coloredlogs.install(
            fmt=self.LOG_FORMAT,  # 使用你的自定义日志格式
            level=self.log_level,
            logger=self.logger,
            field_styles=dict(
                asctime=dict(color="yellow", bold=True),
                name=dict(color="blue", bold=True),
                filename=dict(color="cyan"),
                lineno=dict(color="green", bold=True),
                levelname=dict(color="magenta", bold=True),
            ),
            level_styles=dict(
                critical=dict(color="red", bold=True),
                error=dict(color="red"),
                warning=dict(color="yellow"),
                info=dict(color="green"),
                debug=dict(color="blue"),
            ),
            reconfigure=True,
        )
        self.logger.propagate = False

    def set_file(
        self,
        file: str,
        mode: str = "a",
        max_bytes: int = 10 * 1024 * 1024,
        backup_count: int = 5,
    ):
        """
        Configure file logging for the logger.

        Args:
            file (str): The name of the log file.
            mode (str, optional): File mode. Use 'a' for append or 'w' for overwrite. Defaults to "a".
            max_bytes (int, optional): Maximum size of a single log file before rotation. Defaults to 10MB.
            backup_count (int, optional): Number of backup files to keep. Defaults to 5.
        """
        if mode not in {"a", "w"}:
            raise BioatInvalidParameterError(
                f"Invalid file mode: {mode}. Choose 'a' (append) or 'w' (overwrite)."
            )
        # Ensure the directory exists
        directory = os.path.dirname(file)
        if directory and not os.path.exists(directory):
            os.makedirs(directory)
            self.logger.debug(f"Created directory for log file: {directory}")

        # Remove existing file handlers to prevent duplication
        self.logger.handlers = [
            h for h in self.logger.handlers if not isinstance(h, logging.FileHandler)
        ]

        # Use FileHandler for overwrite mode or RotatingFileHandler for append/rotation
        if mode == "w" or max_bytes <= 0:
            # Simple overwrite mode
            file_handler = logging.FileHandler(file, mode=mode)
        else:
            # Rotating mode
            file_handler = logging.handlers.RotatingFileHandler(
                file, mode=mode, maxBytes=max_bytes, backupCount=backup_count
            )

        # Set file log format
        file_handler.setFormatter(logging.Formatter(self.LOG_FORMAT))
        self.logger.addHandler(file_handler)
        self.logger.propagate = False

    def add_stream_handler(self):
        """
        Add a stream handler to the logger for writing logs to the console.
        """
        # 检查是否已有 StreamHandler
        if not any(isinstance(h, logging.StreamHandler) for h in self.logger.handlers):
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(logging.Formatter(self.LOG_FORMAT))
            self.logger.addHandler(stream_handler)
            self.logger.propagate = False


if __name__ == "__main__":
    # 初始化 LoggerManager
    lm = LoggerManager(log_level="debug")

    print("=== 基础日志测试 ===")
    lm.logger.debug("Debug message: Basic logging test.")
    lm.logger.info("Info message: Basic logging test.")
    lm.logger.warning("Warning message: Basic logging test.")
    lm.logger.error("Error message: Basic logging test.")
    lm.logger.critical("Critical message: Basic logging test.")

    print("\n=== 配置文件日志（追加模式） ===")
    lm.set_file("append_test.log", mode="a")
    lm.logger.info("Info message: Appending to 'append_test.log'.")
    lm.logger.warning("Warning message: Appending to 'append_test.log'.")
    lm.logger.error("Error message: Appending to 'append_test.log'.")

    print("\n=== 配置文件日志（覆盖模式） ===")
    lm.set_file("overwrite_test.log", mode="w")
    lm.logger.info("Info message: Overwriting 'overwrite_test.log'.")
    lm.logger.warning("Warning message: Overwriting 'overwrite_test.log'.")
    lm.logger.error("Error message: Overwriting 'overwrite_test.log'.")

    print("\n=== 配置文件日志（追加到不同文件） ===")
    # 追加另一个文件的同时不影响其他日志的写出
    lm.set_file("different_file.log", mode="a")
    lm.logger.info("Info message: Appending to 'different_file.log'.")
    lm.logger.warning("Warning message: Appending to 'different_file.log'.")
    lm.logger.error("Error message: Appending to 'different_file.log'.")

    print("\n=== 动态更新日志级别 ===")
    lm.set_level("WARNING")
    lm.logger.debug("Debug message: Should not appear (level=WARNING).")
    lm.logger.info("Info message: Should not appear (level=WARNING).")
    lm.logger.warning("Warning message: Should appear (level=WARNING).")
    lm.logger.error("Error message: Should appear (level=WARNING).")

    print("\n=== 动态更新类名和函数名 ===")
    lm.set_level("debug")
    lm.set_names(cls_name="TestClass", func_name="test_function")
    lm.logger.info("Info message: Logging with class and function names.")

    print("\n=== 多次调用文件日志写入 ===")
    lm.set_file("multi_file1.log", mode="w")
    lm.logger.info("First file: multi_file1.log.")
    lm.set_file("multi_file2.log", mode="w")
    lm.logger.info("Second file: multi_file2.log.")
    lm.set_file("multi_file3.log", mode="w")
    lm.logger.info("Third file: multi_file3.log.")

    print("\n=== 测试不合法的日志级别 ===")
    try:
        lm.set_level("fake_level")
    except BioatInvalidParameterError as err:
        print("Should raise FbtValueError")
        print(err)

    print("\n=== 测试重复调用 add_stream_handler ===")
    lm.add_stream_handler()  # 第一次调用
    lm.logger.info("First stream handler added.")
    lm.add_stream_handler()  # 第二次调用，不应重复添加
    lm.logger.info("No duplicate stream handler should be added.")

    print("\n=== 测试自动创建目录 ===")
    lm.set_file("logs/test_folder/append_test.log", mode="a")
    lm.logger.info(
        "This message should be written to 'append_test.log' inside 'logs/test_folder'."
    )



__all__ = ["is_jupyter_environment"]


def is_jupyter_environment() -> bool:
    """Check if the current environment is Jupyter Notebook or JupyterLab."""
    try:
        from IPython import get_ipython

        shell = get_ipython()
        if shell is None:
            return False  # Not in an IPython environment
        # Jupyter Notebook or JupyterLab environment
        if shell.__class__.__name__ in {"ZMQInteractiveShell", "Shell"}:
            return True
        return False
    except ImportError:
        return False  # IPython is not installed, not in Jupyter



import gzip
import math
import os
from io import StringIO
from typing import List

import Bio
import numpy as np
import py3Dmol
from Bio import SeqIO
from Bio.PDB import PDBIO, MMCIFParser, PDBParser, Superimposer
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from bioat.exceptions import (
    BioatFileFormatError,
    BioatFileNotFoundError,
    BioatInvalidParameterError,
)
from bioat.lib.libalignment import (
    get_aligned_seq,
    get_best_alignment,
    instantiate_pairwise_aligner,
)
from bioat.lib.libcolor import map_colors_between_two
from bioat.lib.libfastx import (
    AMINO_ACIDS_1CODE,
    AMINO_ACIDS_3CODE,
    AMINO_ACIDS_3CODE_EXTEND,
)
from bioat.lib.libpath import is_path
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.lib.libpdb")
lm.set_level("DEBUG")


def structure_to_string(structure: Bio.PDB.Structure.Structure):
    """
    将 Bio.PDB.Structure.Structure 对象转换为 PDB 格式的字符串。

    参数:
        structure (Bio.PDB.Structure.Structure): 待转换的结构对象。

    返回:
        str: PDB 格式的字符串。
    """
    pdb_io = PDBIO()
    pdb_io.set_structure(structure)

    # 使用 StringIO 保存到内存中的字符串
    pdb_string = StringIO()
    pdb_io.save(pdb_string)

    # 返回字符串
    return pdb_string.getvalue()


# 对齐函数
def align_cut2ref(
    ref: str | Bio.PDB.Structure.Structure,
    cut: str | Bio.PDB.Structure.Structure,
    gap_indices: list[int],
    label1="ref",
    label2="cut",
):
    """Align pdb to ref using the CA atoms.

    Args:
        ref (str or Bio.PDB.Structure.Structure): Path to the PDB file of the ref structure.
        pdb (str or Bio.PDB.Structure.Structure): Path to the PDB file of the pdb structure.
        gap_indices (list[int]): List of indices of gaps in the cut structure.
        label1 (str, optional): ref label. Defaults to "structure1".
        label2 (str, optional): pdb label. Defaults to "structure2".
        log_level (str, optional): Log level. Defaults to "WARNING".

    Returns:
        Bio.PDB.Structure.Structure: The aligned pdb structure.
    """
    # 读取 PDB 文件
    parser = PDBParser(QUIET=True)

    ref = parser.get_structure(label1, ref) if isinstance(ref, str) else ref
    cut = parser.get_structure(label2, cut) if isinstance(cut, str) else cut

    assert (
        isinstance(ref, Bio.PDB.Structure.Structure)
        and isinstance(cut, Bio.PDB.Structure.Structure)
    ), "Invalid input format. ref and cut must be both str or Bio.PDB.Structure.Structure"

    # 选择对齐的原子 (e.g., CA)
    atoms1 = [atom for atom in ref.get_atoms() if atom.get_name() == "CA"]
    atoms1 = [atoms1[i] for i in range(len(atoms1)) if i not in gap_indices]
    atoms2 = [atom for atom in cut.get_atoms() if atom.get_name() == "CA"]

    assert len(atoms1) == len(
        atoms2
    ), "The number of CA atoms in ref and cut are not equal."
    # 使用 Biopython 的 Superimposer 进行对齐
    super_imposer = Superimposer()
    super_imposer.set_atoms(atoms1, atoms2)
    super_imposer.apply(cut.get_atoms())  # 修改 pdb 的坐标
    rmsd = super_imposer.rms

    return {
        label1: ref,
        label2: cut,
        "RMSD": rmsd,
    }

def _load_seq(seq, label=None):
    if isinstance(seq, str) and is_path(seq):
        seq = next(SeqIO.parse(seq, "fasta"))
        label = seq.id if label is None else label
        seq = seq.seq
    elif isinstance(seq, Seq):
        label = "no_name" if label is None else label
    else:
        raise BioatInvalidParameterError(f"Invalid sequence format. seq: {seq}")
    return seq, label


def show_ref_cut(
    ref_seq: str | Seq,
    ref_pdb: str | Bio.PDB.Structure.Structure,
    cut_seq: List[str | Seq] | str | Seq | None = None,
    cut_pdb: List[str | Bio.PDB.Structure.Structure]
    | str
    | Bio.PDB.Structure.Structure
    | None = None,
    cut_labels: List[str] | str | None = None,
    ref_color: str = "red",
    ref_color_base: str | None = None,
    ref_value_dict: dict | None = None,
    cut_color="lightgray",
    gap_color="purple",
    ref_style="cartoon",
    cut_style="cartoon",
    gap_style="cartoon",
    ref_value_random: bool = False,
    output_fig: str | None = None,
    col: int = 3,
    scale: float = 1.0,
    annotate: bool = True,
    log_level="WARNING",
):
    """
    Visualizes the alignment of sequences and highlights changes in PDB structures using py3Dmol.

    Args:
        ref_seq (str or Seq): Amino acid sequence content for the ref protein.
        ref_pdb (str or Bio.PDB.Structure.Structure): Path to the PDB file of the reference structure.
        cut_seq (str, Seq or None, optional): Amino acid sequence content for the cut protein.
        cut_pdb (str, Bio.PDB.Structure.Structure or None, optional): Path to the PDB file of the cut structure.
        cut_labels (str or None, optional): Label for the cut proteins. If None, the label will be set to "cut".
        ref_color (str, optional): Color for reference residues.
        ref_color_base (str or None, optional): ref_color_base will be used as base color, and ref_color will be used as target color. If None, do not apply color mapping. Defaults to None.
        ref_value_dict (dict or None, optional): A dictionary of values for the ref color map, it will be normalized to the range of [0 - 1]. If None, all residues will be colored with the same color. e.g. ref_value_dict = {'V_0': 0.4177215189873418, 'S_1': 0.8185654008438819, 'K_2': 0.9915611814345991, 'G_3': 0.42616033755274263, ...}
        cut_color (str, optional): Color for cut residues.
        gap_color (str, optional): Color for gaps or removed residues.
        ref_style (str, optional): "stick", "sphere", "cartoon", or "line"
        cut_style (str, optional): "stick", "sphere", "cartoon", or "line"
        gap_style (str, optional): "stick", "sphere", "cartoon", or "line"
        ref_value_random (bool, optional): If True, ref_value_dict will be randomly generated. Defaults to False.
        output_fig (str or None, optional): Output figure file path. If None, the figure will not be saved in html format. Defaults to None.
        col (int, optional): Number of columns for the visualization. Defaults to 3.
        log_level (str, optional): Log level. Defaults to "WARNING".
    """
    lm.set_names(func_name="show_ref_cut")
    lm.set_level(log_level)

    aligner = instantiate_pairwise_aligner(
        scoring_match=5,
        penalty_mismatch=-1,
        penalty_gap_open=-1,
        penalty_gap_extension=0,
        penalty_query_left_gap_score=0,
        penalty_query_right_gap_score=0,
        log_level="WARNING",
    )

    # fix color mapping
    assert (ref_color_base is None and ref_value_dict is None) or (
        isinstance(ref_color_base, str) and isinstance(ref_value_dict, dict)
    ), "ref_color_base and ref_value_dict must be both None or both not None"

    ref_seq, ref_label = _load_seq(ref_seq, "ref")

    # load ref_pdb
    parser = PDBParser(QUIET=True)
    ref_pdb = (
        ref_pdb
        if isinstance(ref_pdb, Bio.PDB.Structure.Structure)
        else parser.get_structure(ref_label, ref_pdb)
    )
    # ref_seq ref_pdb ref_label

    if cut_seq:  # not None
        if not isinstance(cut_seq, list):
            cut_seqs = [cut_seq]
        else:
            cut_seqs = cut_seq
        if not isinstance(cut_pdb, list):
            cut_pdbs = [cut_pdb]
        else:
            cut_pdbs = cut_pdb
        ref_seqs_fix = []
        ref_pdbs_fix = []
        ref_labels = []
        cut_seqs_fix = []
        cut_pdbs_fix = []
        cut_labels_fix = []
        gap_indices = []
        rmsds = []
        alignment_dicts = []

        for idx, cut_seq in enumerate(cut_seqs):
            ref_seqs_fix.append(ref_seq)
            ref_labels.append(ref_label)
            # 执行Sequence Alignment并得到gaps的位置
            cut_seq, cut_label = _load_seq(cut_seq, "cut")
            cut_seqs_fix.append(cut_seq)
            if cut_labels:
                if isinstance(cut_labels, list):
                    cut_labels_fix.append(cut_labels[idx])
                elif isinstance(cut_labels, str):
                    cut_labels_fix.append(cut_labels)
            else:
                cut_labels_fix.append(cut_label)
            # 获得gaps的位置
            # Perform alignment
            alignment = get_best_alignment(
                seq_a=ref_seq, seq_b=cut_seq, aligner=aligner, consider_strand=False
            )
            alignment_dict = get_aligned_seq(alignment, reverse=False, letn_match=False)
            alignment_dicts.append(alignment_dict)
            gap_indices.append(
                [i for i, n in enumerate(alignment_dict["target_seq"]) if n == "-"]
            )

            cut_pdb = cut_pdbs[idx]

            if cut_pdb:
                # 执行PDB Alignment并得对齐后的PDB文件和RMSD
                res = align_cut2ref(
                    ref_pdb, cut_pdb, gap_indices[idx], ref_label, cut_label
                )
                ref_pdbs_fix.append(res[ref_label])
                cut_pdbs_fix.append(res[cut_label])
                rmsds.append(res["RMSD"])
            else:
                ref_pdbs_fix.append(ref_pdb)
                cut_pdbs_fix.append(None)
                rmsds.append(None)
    else:
        lm.logger.info(
            "Cut sequence is not provided, skip alignment and annotation for gaps, and just show ref visualization."
        )
        ref_seqs_fix = [ref_seq]
        ref_pdbs_fix = [ref_pdb]
        ref_labels = [ref_label]
        cut_seqs_fix = [None]
        cut_pdbs_fix = [None]
        cut_labels_fix = [None]
        gap_indices = [None]
        rmsds = [None]
        alignment_dicts = [None]

    assert all(
        [isinstance(ref_pdb, Bio.PDB.Structure.Structure) for ref_pdb in ref_pdbs_fix]
    ), "Invalid input format. ref_pdb must be Bio.PDB.Structure"

    if cut_pdbs_fix[0]:
        assert all(
            [
                isinstance(cut_pdb, Bio.PDB.Structure.Structure)
                for cut_pdb in cut_pdbs_fix
            ]
        ), "Invalid input format. cut_pdb must be Bio.PDB.Structure"

    n = len(ref_seqs_fix)
    col = min(n, col)
    row = math.ceil(n / col)

    view = py3Dmol.view(
        width=600 * col * scale,
        height=600 * row * scale,
        viewergrid=(row, col) if n > 1 else None,
        linked=True,  # 是否同步网格间的分子运动
    )
    lm.logger.debug(f"viewergrid: ({row}, {col})")
    viewers = [(i // col, i % col) for i in range(n)]
    lm.logger.debug(f"viewers = {viewers}")

    for i, viewer in enumerate(viewers):
        lm.logger.debug(f"viewer = {viewer}")
        _add_one_submodel(
            view=view,
            ref_seq=ref_seqs_fix[i],
            ref_pdb=ref_pdbs_fix[i],
            ref_label=ref_labels[i],
            cut_seq=cut_seqs_fix[i],
            cut_pdb=cut_pdbs_fix[i],
            cut_label=cut_labels_fix[i],
            ref_color=ref_color,
            ref_color_base=ref_color_base,
            ref_value_dict=ref_value_dict,
            ref_value_random=ref_value_random,
            cut_color=cut_color,
            gap_color=gap_color,
            gap_indices=gap_indices[i],
            ref_style=ref_style,
            cut_style=cut_style,
            gap_style=gap_style,
            alignment_dict=alignment_dicts[i],
            rmsd=rmsds[i],
            annotate=annotate,
            viewer=viewer,
        )

    if output_fig:
        assert output_fig.endswith(".html"), "Output file must have a '.html' extension"
        # 保存为 SVG 文件
        with open(output_fig, "w") as f:
            f.write(view._make_html())
    # display(HTML(view._make_html()))
    # view.show()
    # 确保所有子视图都已完成模型添加和样式设置
    # view.render()  # 显式调用渲染
    # view.show()
    # return view.png()
    # view.zoomTo()
    return view.show()


def _add_one_submodel(
    view,
    ref_seq,
    ref_pdb,
    ref_label,
    cut_seq,
    cut_pdb,
    cut_label,
    ref_color,
    ref_color_base,
    ref_value_dict,
    ref_value_random,
    cut_color,
    gap_color,
    gap_indices,
    ref_style,
    cut_style,
    gap_style,
    alignment_dict,
    rmsd,
    annotate,
    viewer,
):
    # for annotation
    label_xpos = -70
    label_ypos = 75
    label_zpos = -30

    colors = None
    if ref_value_dict:
        values = []
        res_names = []
        res_idxes = []
        for k, v in ref_value_dict.items():
            res_name, res_idx = k.split("_")
            assert res_name in AMINO_ACIDS_1CODE, f"Invalid residue name: {res_name}"
            res_idx = int(res_idx)
            values.append(v)
            res_names.append(res_name)
            res_idxes.append(res_idx)
        if not ref_value_random:
            colors = map_colors_between_two(ref_color_base, ref_color, values)
        else:
            colors = map_colors_between_two(
                ref_color_base, ref_color, np.random.rand(len(values))
            )

    view.setViewStyle({"style": "outline", "color": "black", "width": 0.1})

    view.addModel(
        structure_to_string(ref_pdb),
        "pdb",
        viewer=viewer,
    )
    view.setStyle(
        {"model": 0},
        {
            ref_style: {
                "arrows": True,
                "tubes": True,
                "style": "oval",
                "color": ref_color,
            },
        },
        viewer=viewer,
    )
    label_ypos -= 5
    view.addLabel(
        f"{ref_label}: {ref_color}",
        {
            "fontColor": ref_color,
            "fontSize": 14,
            "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
        },
        viewer=viewer,
    )

    # 标记ref 值的颜色
    if ref_color_base is not None:
        for i in range(len(colors)):
            idx = res_idxes[i]
            res = res_names[i]
            color = colors[i]
            view.addStyle(
                {"model": 0, "resi": idx + 1, "resn": AMINO_ACIDS_1CODE[res]},
                {ref_style: {"color": color, "radius": 0.5}},
                viewer=viewer,
            )
        if annotate:
            label_ypos -= 5
            view.addLabel(
                f"{ref_label} values: {ref_color_base} (0) to {ref_color} (1)",
                {
                    "fontColor": ref_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )
    # 标记gaps的颜色
    if gap_indices:
        for idx, res in enumerate(alignment_dict["reference_seq"]):
            # 标记 gap位置的颜色
            if idx in gap_indices:
                view.addStyle(
                    {"model": 0, "resi": idx + 1, "resn": AMINO_ACIDS_1CODE[res]},
                    {gap_style: {"color": gap_color, "radius": 0.5}},
                    viewer=viewer,
                )
        if annotate:
            label_ypos -= 5
            view.addLabel(
                f"gaps: {gap_color}",
                {
                    "fontColor": gap_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )
    if cut_pdb and rmsd:
        view.addModel(structure_to_string(cut_pdb), "pdb", viewer=viewer)
        view.setStyle(
            {"model": 1},
            {
                cut_style: {
                    "arrows": True,
                    "tubes": True,
                    "style": "oval",
                    "color": cut_color,
                },
            },
            viewer=viewer,
        )
        if annotate:
            label_ypos -= 5
            view.addLabel(
                f"{cut_label}: {cut_color}",
                {
                    "fontColor": cut_color,
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )
            label_ypos -= 5
            view.addLabel(
                f"RMSD: {rmsd:.2f}",
                {
                    "fontColor": "white",
                    "fontSize": 14,
                    "position": {"x": label_xpos, "y": label_ypos, "z": label_zpos},
                },
                viewer=viewer,
            )

    # 聚焦显示
    # Finalize the view
    view.zoomTo(viewer=viewer)
    # view.render(viewer=viewer)


def pdb2fasta(pdb_file, output_fasta, log_level="DEBUG"):
    """Converts a PDB / CIF file to a FASTA file.

        This function processes the provided PDB file and extracts protein, DNA,
        RNA sequences, and other molecules appropriately to create a FASTA file.
    e   1. Proteins:The protein sequence for each chain will be extracted as Chain X Protein.
        2. DNA and RNA: Bases for DNA (A, T, G, C) will be saved as Chain X DNA and bases for RNA (A, U, G, C) will be saved as Chain X RNA.
        3. Other molecules: Any unrecognized molecules (e.g., ions, modified molecules) will be labeled as [residue] and stored as Chain X Other molecules.
        4. Multi-chain complexes: The program supports multi-chain structures in complexes, and the content of each chain will be recorded separately.

        Args:
            pdb_file (str): input file path.
            output_fasta (str | None, optional): output file path. If None, the output file will be named as the basename of the input file with a ".fa" extension. Defaults to None.
            log_level (str, optional): log level. Defaults to "WARNING".
    """
    lm.set_names(func_name="pdb2fasta")
    lm.set_level(log_level)

    # 设置输出文件名
    if output_fasta is None:
        if pdb_file.endswith(".gz"):
            idx = ".".join(os.path.basename(pdb_file).split(".")[:-2])
        else:
            idx = ".".join(os.path.basename(pdb_file).split(".")[:-1])
        output_fasta = f"{idx}.fa" if idx else "output.fa"
        lm.logger.debug(f"Output is not defined, set output to {output_fasta}")
    else:
        assert output_fasta.endswith(".fa"), "Output file must have a '.fa' extension"
        idx = ".".join(os.path.basename(output_fasta).split(".fa")[:-1])

    # 检查输入文件是否存在
    if not os.path.exists(pdb_file):
        raise BioatFileNotFoundError(f"Input file not found: {pdb_file}")

    # 检测文件格式并选择解析器
    if pdb_file.endswith(".cif") or pdb_file.endswith(".cif.gz"):
        parser = MMCIFParser(QUIET=True)
    elif pdb_file.endswith(".pdb") or pdb_file.endswith(".pdb.gz"):
        parser = PDBParser(QUIET=True)
    else:
        raise BioatFileFormatError("Only .pdb and .cif files are allowed.")

    # 打开文件，支持 gzip 格式
    input_handler = (
        gzip.open(pdb_file, "rt") if pdb_file.endswith(".gz") else open(pdb_file, "rt")
    )
    structure = parser.get_structure(idx, input_handler)

    records = []

    # 遍历结构中的所有模型和链，提取序列
    for model in structure:
        for chain in model:
            lm.logger.debug(f"processing chain {chain.id}")
            # 找出当前链中所有氨基酸的编号范围
            residue_ids = [residue.id[1] for residue in chain]

            if not residue_ids:
                continue

            min_res_id, max_res_id = min(residue_ids), max(residue_ids)
            # 分别初始化序列列表，使用 "-" 占位符填充缺失位置
            protein_seq = ["-"] * (max_res_id - min_res_id + 1)
            dna_seq = ["-"] * (max_res_id - min_res_id + 1)
            rna_seq = ["-"] * (max_res_id - min_res_id + 1)
            other_seq = ["-"] * (max_res_id - min_res_id + 1)

            for residue in chain:
                residue_name = residue.resname.strip().upper()  # 标准化残基名称
                residue_id = residue.id[1] - min_res_id  # 计算相对于最小编号的偏移

                if residue_name in AMINO_ACIDS_3CODE:
                    # 蛋白质序列，填充氨基酸到相应位置
                    protein_seq[residue_id] = AMINO_ACIDS_3CODE[residue_name]
                elif residue_name in AMINO_ACIDS_3CODE_EXTEND:
                    # 蛋白质序列，填充氨基酸到相应位置
                    protein_seq[residue_id] = AMINO_ACIDS_3CODE_EXTEND[residue_name]
                elif residue_name in ("DA", "DT", "DG", "DC"):
                    # DNA 序列，填充 A, T, G, C
                    dna_seq[residue_id] = residue_name[1]
                elif residue_name in ("A", "U", "G", "C"):
                    # RNA 碱基
                    rna_seq[residue_id] = residue_name
                else:
                    # 其他分子类型，用 [] 标记
                    other_seq[residue_id] = f"[{residue_name}]"

            # 根据链类型创建 SeqRecord，并添加到 records 列表
            if any(res != "-" for res in protein_seq):
                records.append(
                    SeqRecord(
                        Seq("".join(protein_seq)),
                        id=f"{structure.id}|Chain_{chain.id}|Protein",
                        description=f"Chain {chain.id} Protein",
                    )
                )
            if any(res != "-" for res in dna_seq):
                records.append(
                    SeqRecord(
                        Seq("".join(dna_seq)),
                        id=f"{structure.id}|Chain_{chain.id}|DNA",
                        description=f"Chain {chain.id} DNA",
                    )
                )
            if any(res != "-" for res in rna_seq):
                records.append(
                    SeqRecord(
                        Seq("".join(rna_seq)),
                        id=f"{structure.id}|Chain_{chain.id}|RNA",
                        description=f"Chain {chain.id} RNA",
                    )
                )
            if any(res != "-" for res in other_seq):
                records.append(
                    SeqRecord(
                        Seq("".join(other_seq)),
                        id=f"{structure.id}|Chain_{chain.id}|UNKNOWN",
                        description=f"Chain {chain.id} Other molecules",
                    )
                )

    # 将所有链的序列保存为 FASTA 文件
    SeqIO.write(records, output_fasta, "fasta")
    input_handler.close()
    lm.logger.info(f"FASTA saved to: {output_fasta}")


if __name__ == "__main__":
    # ref_seq
    pass



"""Main module of the bioat package.

This module serves as the command-line interface (CLI) for the bioat package.
It uses the Fire library to automatically generate help and command-line parsing capabilities.

Usage:
    To see the available commands and options, run:
        python -m bioat -h
    or simply:
        bioat -h

Attributes:
    calculator (Cli): An instance of the Cli class that handles command-line operations.

Functions:
    main() -> int: The entry point of the module that sets up and runs the CLI.
"""

import fire

from bioat.cli import Cli


def main() -> None:
    calculator = Cli()
    # print in shell stdout instead of view in `less`
    fire.core.Display = lambda lines, out: print(*lines, file=out)
    #
    fire.Fire(calculator, name='bioat')


if __name__ == '__main__':
    print(type(main()))



import os
import time

import psutil

from bioat.exceptions import BioatInvalidParameterError
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.devtools")


def _elapsed_since(start):
    return time.strftime("%Mm:%Ss", time.gmtime(time.time() - start))


def _get_process_memory():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


def profile(num_iterations=1):
    """
    Decorator for calculating the runtime of a function, over multiple iterations.

    :param num_iterations: times to iter, defaults to 1
    :type num_iterations: int, optional
    """

    def inner_decorator(func):
        def wrapper(*args, **kwargs):
            if not isinstance(num_iterations, int) or num_iterations < 1:
                raise BioatInvalidParameterError(
                    "num_iterations must be a positive integer"
                )
            lm.set_names(func_name=func.__name__)
            lm.set_level("INFO")

            total_time = 0
            start_time = 0
            result = None

            mem_before = _get_process_memory()
            for _ in range(num_iterations):
                start_time = time.time()
                result = func(*args, **kwargs)
                elapsed_time = time.time() - start_time
                total_time += elapsed_time

            mem_after = _get_process_memory()
            mem_diff_mb = (mem_after - mem_before) / (1024**2)  # 转换为 MB

            avg_time = total_time / num_iterations
            elapsed_time_formatted = _elapsed_since(start_time)

            text = (
                f"\n[{func.__name__}] performance profiling:\n"
                f"\t↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓\n"
                f"\tIteration num: {num_iterations}x\n"
                f"\tTotal    time: {total_time:.9f}s ({elapsed_time_formatted})\n"
                f"\tAverage  time: {avg_time:.9f}s\n"
                f"\tMemory   diff: {mem_diff_mb:.2f} MB"
            )
            lm.logger.info(text)

            return result

        return wrapper

    return inner_decorator


"""
crisprtools.py

This module provides a toolbox for mining CRISPR-related sequences in metagenomic data.
It includes functionality for identifying Cas candidates associated with CRISPR loci
and for annotating Cas13 candidates from protein sequences. The main class, `CrisprTools`,
provides methods for calling external executables such as Prodigal and Pilercr for protein
prediction and CRISPR identification.

Classes:
    CrisprTools: A class containing methods for identifying Cas candidates and Cas13
                 candidates from genomic data.

Methods:
    cas_finder: A method for de novo annotation of Cas candidates from CRISPR loci,
                utilizing input fasta files and producing various output files.
    cas13_finder: A method for the annotation of Cas13 candidates from protein
                  fasta files.

Usage:
    To use this module, create an instance of `CrisprTools` and call the methods
    `cas_finder` or `cas13_finder` with the appropriate parameters to perform
    CRISPR analysis on your dataset.
"""

from bioat.lib.libfastx import cas13_finder, cas_finder
from bioat.lib.libpath import check_executable


class CrisprTools:
    """CRISPR mining toolbox.

    This class provides methods for performing CRISPR analysis on datasets, including
    finding Cas proteins and CRISPR sequences.

    Attributes:
        None
    """

    def __init__(self):
        pass

    def cas_finder(
        self,
        input_fa,
        output_faa=None,
        output_contig_fa=None,
        output_crispr_info_tab=None,
        lmin=3000,  # 3001 in Nature Methods paper
        lmax=None,
        extend=10_000,
        temp_dir=None,
        prodigal=None,
        prodigal_mode="meta",
        pilercr=None,
        rm_temp=True,
        log_level="INFO",
    ):
        """De novo annotation for Cas candidates from neighbors of CRISPR loci.

        Args:
            input_fa (str): Path to the input metagenome fasta file containing many contigs.
            output_faa (str, optional): Path to save the de novo annotated Cas candidates.
            output_contig_fa (str, optional): Path to save the whole contigs of de novo annotated Cas candidates.
            output_crispr_info_tab (str, optional): Path to save the de novo annotated CRISPR info table (CSV format).
            lmin (int, optional): Minimum length for a contig. Defaults to 3000.
            lmax (int, optional): Maximum length for a contig. Defaults to None.
            extend (int, optional): Distance over which proteins are considered from the start/end of the CRISPR loci. Defaults to 10000.
            temp_dir (str, optional): Directory to store temporary files. Defaults to None.
            prodigal (str, optional): Path to the Prodigal executable. Defaults to None.
            prodigal_mode (str, optional): Mode for Prodigal annotation. Can be "meta" or "single". Defaults to "meta".
            pilercr (str, optional): Path to the Pilercr executable. Defaults to None.
            rm_temp (bool, optional): If False, temporary files will be kept. Defaults to True.
            log_level (str, optional): Logging level; set to "DEBUG" to see logs from prodigal and pilercr. Defaults to "INFO".
        """
        check_executable(name=prodigal)
        check_executable(name=pilercr)
        cas_finder(
            input_fa=input_fa,
            output_faa=output_faa,
            output_contig_fa=output_contig_fa,
            output_crispr_info_tab=output_crispr_info_tab,
            lmin=lmin,
            lmax=lmax,
            extend=extend,
            temp_dir=temp_dir,
            prodigal=prodigal,
            prodigal_mode=prodigal_mode,
            pilercr=pilercr,
            rm_temp=rm_temp,
            log_level=log_level,
        )

    # def cas12_finder(self):
    #     """Cas12 mining toolbox.
    #
    #     Under development!
    #     """
    #     pass

    def cas13_finder(
        self, input_faa, output_faa=None, lmin=200, lmax=1500, log_level="INFO"
    ):
        """De novo annotation for Cas13 candidates from proteins.faa.

        Args:
            input_faa (str): The input file containing Cas candidates in .faa format.
            output_faa (str, optional): The output file for Cas13 candidates in .faa format. Defaults to None.
            lmin (int, optional): Minimum length for a Cas candidate. Defaults to 200.
            lmax (int, optional): Maximum length for a Cas candidate. Defaults to 1500.
            log_level (str, optional): The logging level. Options are 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'. Defaults to 'INFO'.
        """
        cas13_finder(
            input_faa=input_faa,
            output_faa=output_faa,
            lmin=lmin,
            lmax=lmax,
            log_level=log_level,
        )


if __name__ == "__main__":
    pass



"""Doc.
Adapt from https://github.com/ponnhide/pyCircos
and thanks for this github user.
"""
import io
import math
import os
import re
import tempfile
import urllib

import Bio
import matplotlib
import matplotlib.patches as mpatches
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from Bio import Phylo, SeqIO

matplotlib.rcParams["figure.max_open_warning"] = 0
matplotlib.rcParams['ps.fonttype'] = 42
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['font.sans-serif'] = ["Arial", "Lucida Sans", "DejaVu Sans", "Lucida Grande", "Verdana"]
matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.size'] = 10.0
matplotlib.rcParams["axes.labelcolor"] = "#000000"
matplotlib.rcParams["axes.linewidth"] = 1.0
matplotlib.rcParams["xtick.major.width"] = 1.0
matplotlib.rcParams["ytick.major.width"] = 1.0
matplotlib.rcParams['xtick.major.pad'] = 6
matplotlib.rcParams['ytick.major.pad'] = 6
matplotlib.rcParams['xtick.major.size'] = 6
matplotlib.rcParams['ytick.major.size'] = 6
DATAPATH = os.path.join(os.path.dirname(__file__), 'circos')


class Garc:
    """Class representation for an arc section in a graphic circle.

    Attributes:
        colorlist (list of str): List of color codes for the arc sections.
        _arcnum (int): Counter for the number of arcs.

    Methods:
        __setitem__(key, item):
            Sets the attribute `key` with the value `item`.
        __getitem__(key):
            Retrieves the value of the attribute `key`.
    """

    colorlist = ["#ff8a80", "#ff80ab", "#ea80fc", "#b388ff", "#8c9eff", "#82b1ff", "#84ffff", "#a7ffeb", "#b9f6ca",
                 "#ccff90", "#f4ff81", "#ffff8d", "#ffe57f", "#ffd180", "#ff9e80", "#bcaaa4", "#eeeeee", "#b0bec5",
                 "#ff5252", "#ff4081", "#e040fb", "#7c4dff", "#536dfe", "#448aff", "#18ffff", "#64ffda", "#69f0ae",
                 "#b2ff59", "#eeff41", "#ffff00", "#ffd740", "#ffab40", "#ff6e40", "#a1887f", "#e0e0e0", "#90a4ae"]
    _arcnum = 0

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __init__(
        self,
        arc_id=None,
        record=None,
        size=1000,
        interspace=3,
        raxis_range=(500, 550),
        facecolor=None,
        edgecolor="#303030",
        linewidth=0.75,
        label=None,
        labelposition=0,
        labelsize=10,
        label_visible=False,
    ):
        """Initializes the Garc object with specified parameters.

        Args:
            arc_id (str, optional):
                Unique identifier for the Garc class object. If None, a unique ID is generated. Default is None.
            record (Bio.SeqRecord or str, optional):
                Bio.SeqRecord object or NCBI accession number for an annotated sequence. Default is None.
            size (int, optional):
                Width of the arc section, default is 1000. Adjusted if record is provided.
            interspace (float, optional):
                Distance angle (degrees) to the adjacent arc section. Default is 3.
            raxis_range (tuple of int, optional):
                Radial axis range for line plotting. Default is (500, 550).
            facecolor (str or tuple, optional):
                Color for filling the arc section. Default is automatically set.
            edgecolor (str or tuple, optional):
                Color for the edge of the filled area. Default is "#303030".
            linewidth (float, optional):
                Width of the edge line. Default is 0.75.
            label (str, optional):
                Label for the arc section. Default is None.
            labelposition (int, optional):
                Relative label height from the center. Default is 0.
            labelsize (int, optional):
                Font size of the label. Default is 10.
            label_visible (bool, optional):
                Determines if the label is visible on the arc section. Default is False.

        Raises:
            ValueError:
                If no match for the provided NCBI accession number is found in `record`.
        """

        self._parental_gcircle = None
        if arc_id == None:
            self.arc_id = str(Garc._arcnum)
        else:
            self.arc_id = arc_id

        if record is None:
            self.record = None
            self.size = size

        elif type(record) == Bio.SeqRecord.SeqRecord:
            self.record = record
            self.size = len(str(self.record.seq))

        elif type(record) == str:
            match = re.fullmatch("[a-zA-Z]{1,2}_?[0-9]{5,6}", record)
            if os.path.exists(record) == True:
                self.record = SeqIO.read(record, format="genbank")

            if match is None:
                raise ValueError("Incorrect value for NCBI accession number.")
            else:
                url = "https://www.ncbi.nlm.nih.gov/sviewer/viewer.cgi?tool=portal&save=file&log$=seqview&db=nuccore&report=gbwithparts&id={}&withparts=on".format(
                    record
                )
            outb = io.BytesIO()
            outs = io.StringIO()
            headers = {
                "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:47.0) Gecko/20100101 Firefox/47.0"
            }
            request = urllib.request.Request(url, headers=headers)

            with urllib.request.urlopen(request) as u:
                outb.write(u.read())
            outs.write(outb.getvalue().decode())

            with tempfile.TemporaryFile(mode="w+") as o:
                content = outs.getvalue()
                o.write(content)
                o.seek(0)
                record = SeqIO.parse(o, "genbank")
                record = next(record)
            self.record = record
            self.size = len(str(self.record.seq))
        else:
            self.record = None
            self.size = size

        if facecolor is None:
            facecolor = Garc.colorlist[Garc._arcnum % len(Garc.colorlist)]
        self.interspace = 2 * np.pi * (interspace / 360)
        self.raxis_range = raxis_range
        self.facecolor = facecolor
        self.edgecolor = edgecolor
        self.linewidth = linewidth

        if label is None:
            self.label = arc_id
        else:
            self.label = label

        self.label_visible = label_visible
        self.labelposition = labelposition
        self.labelsize = labelsize
        Garc._arcnum += 1

    def calc_density(self, positions, window_size=1000):
        """Calculates density values in a sliding window based on given positions.

        Args:
            positions (list of int or tuple): A list of x-coordinate values
                or a tuple containing two x-coordinate values. Each value
                should be in the range of 0 to the size of the Garc object.
            window_size (int, optional): The size of the sliding window.
                Defaults to 1000.

        Raises:
            ValueError: If an inappropriate value or values are provided
                for positions.

        Returns:
            list: A list containing the calculated density values.
        """

        densities = []
        positions.sort()
        for i in range(0, self.size, window_size):
            source = tuple(range(i, i + window_size))
            amount = 0
            for pos in positions:
                if type(pos) == int:
                    if pos in source:
                        amount += 1
                elif type(pos) == tuple:
                    if pos[0] <= source[-1] and pos[1] >= source[0]:
                        amount += 1
                else:
                    raise ValueError("List elements should be int type or tuple consisting of two int values")
            densities.append(amount)

        source = tuple(range(i, self.size))
        amount = 0
        for pos in positions:
            if type(pos) == int:
                if pos in source:
                    amount += 1
            elif type(pos) == tuple:
                if pos[0] <= source[-1] and pos[1] >= source[0]:
                    amount += 1
            else:
                raise ValueError("List elements should be int type or tuple consisting of two int values")
        densities.append(amount * ((self.size - i) / window_size))
        return densities

    def calc_nnratio(self, n1="G", n2="C", window_size=1000, step_size=None):
        """Calculates the ratio of nucleotide base frequencies for two specified bases over multiple windows along a sequence.

        Args:
            n1 (str): The first nucleotide base to be compared.
                Must be one of "ATGC". Default is "G".
            n2 (str): The second nucleotide base to be compared.
                Must be one of "ATGC". Default is "C".
            window_size (int): Size of the sliding window.
                Default is 1000.
            step_size (int, optional): Size of the sliding step.
                If not provided, defaults to window_size.

        Raises:
            ValueError: If no record is provided.

        Returns:
            np.array: An array containing the computed ratios of n1 to n2 for each window.
        """

        if self.record is None:
            raise ValueError("self.record is None, please specify record value")

        if step_size is None:
            step_size = window_size

        seq = str(self.record.seq)
        gc_amounts = []
        for i in range(0, len(seq), step_size):
            if n2 is None:
                gc_amount = (
                    seq[i : i + window_size].upper().count(n1) * 1.0 / window_size
                )
            else:
                gc_amount = (
                    (
                        seq[i : i + window_size].upper().count(n1)
                        + seq[i : i + window_size].upper().count(n2)
                    )
                    * 1.0
                    / window_size
                )
            gc_amounts.append(gc_amount)
        if n2 is None:
            gc_amounts.append(seq[i:].upper().count(n1) * 1.0 / (len(seq) - i))
        else:
            gc_amounts.append(
                (seq[i:].upper().count(n1) + seq[i : i + window_size].upper().count(n2))
                * 1.0
                / (len(seq) - i)
            )

        self["{}{}_ratio".format(n1, n2)] = gc_amounts
        gc_amounts = np.array(gc_amounts)
        return gc_amounts

    def calc_nnskew(self, n1="G", n2="C", window_size=1000, step_size=None):
        """
        Calculates n1,n2 skew (n1-n2)/(n1+n2) for multiple windows along
        the sequence.

        Args:
            n1 (str, optional): The first of the two nucleotide bases to be compared.
                The default is "G".
            n2 (str, optional): The second of the two nucleotide bases to be compared.
                The default is "C".
            window_size (int, optional): Size of the sliding window. The default is 1000.
            step_size (int, optional): Size of the sliding step. The default is window_size.

        Raises:
            ValueError: If no record is provided, will return an error.

        Returns:
            np.array: An array of the skews computed by this method.
        """

        # (G-C)/(G+C)
        if self.record is None:
            raise ValueError("self.record is None, please specify record value")

        if step_size is None:
            step_size = window_size

        seq = str(self.record.seq)
        gc_skews = []
        for i in range(0, len(seq), step_size):
            gc_skew = (seq[i:i + window_size].upper().count(n1) - seq[i:i + window_size].upper().count(n2)) * 1.0 / (
                    seq[i:i + window_size].upper().count(n1) + seq[i:i + window_size].upper().count(n2)) * 1.0
            gc_skews.append(gc_skew)

        gc_skews.append((seq[i:].upper().count(n1) - seq[i:].upper().count(n2)) * 1.0 / (
                seq[i:].upper().count(n1) + seq[i:].upper().count(n2)) * 1.0)
        self["{}{}_skew".format(n1, n2)] = gc_skews
        gc_skews = np.array(gc_skews)
        return gc_skews


class Gcircle:
    """
    A Gcircle class object provides a circle whose diameter is 1000 (a.u.) as a 
    drawing space. Any graph (line plot, scatter plot, barplot, heatmap, and chordplot) 
    can be placed on the space by specifying the raxis_range (from 0 to 1000) and 
    the corresponding Garc class object.
    """
    colors = ["#f44336", "#e91e63", "#9c27b0", "#673ab7", "#3f51b5", "#2196f3", "#00bcd4", "#009688", "#4caf50",
              "#8bc34a", "#cddc39", "#ffeb3b", "#ffc107", "#ff9800", "#ff5722", "#795548", "#9e9e9e", "#607d8b"]
    # colors = ["#4E79A7","#F2BE2B","#E15759","#76B7B2","#59A14F","#EDC948","#B07AA1","#FF9DA7","#9C755F","#BAB0AC"]
    cmaps = [plt.cm.Reds, plt.cm.Blues, plt.cm.Greens, plt.cm.Greys]

    def __getattr__(self, name):
        if name == "garc_dict":
            return self._garc_dict

    def __init__(self, fig=None, figsize=None):
        """
        Initializes the circular map.

        Args:
            fig (matplotlib.pyplot.Figure, optional): A Matplotlib Figure object. If not provided, a new figure will be created.
            figsize (tuple, optional): A tuple specifying the size of the figure for the circular map, e.g., (width, height).

        """

        self._garc_dict = {}
        if fig is None:
            if figsize is None:
                figsize = (8, 8)
            self.figure = plt.figure(figsize=figsize)
            self.fig_is_ext = False
        else:
            if figsize is None:
                figsize = (6, 6)
            self.figure = fig
            self.fig_is_ext = True
        self.figsize = figsize
        self.color_cycle = 0

    def add_garc(self, garc):
        """Add a new Garc class object to the garc_dict.

        Args:
            garc (Garc): The Garc class object to be added.

        Returns:
            None
        """

        self._garc_dict[garc.arc_id] = garc

    def set_garcs(self, start=0, end=360):
        """Visualize arc rectangles for Garc class objects in .garc_dict.

        This method draws the arc rectangles of the Garc class objects in the drawing space.
        After executing this method, no new Garc class objects can be added to garc_dict,
        and a figure parameter representing a matplotlib.pyplot.figure object will be created
        in the Gcircle object.

        Args:
            start (int, optional):
                Start angle of the circos plot. The range is -360 to 360.
                Defaults to 0.
            end (int, optional):
                End angle of the circos plot. The range is -360 to 360.
                Defaults to 360.

        Returns:
            None
        """

        sum_length = sum(
            list(
                map(lambda x: self._garc_dict[x]["size"], list(self._garc_dict.keys()))
            )
        )
        sum_interspace = sum(
            list(
                map(
                    lambda x: self._garc_dict[x]["interspace"],
                    list(self._garc_dict.keys()),
                )
            )
        )
        start = 2 * np.pi * start / 360
        end = (2 * np.pi * end / 360) - sum_interspace

        s = 0
        sum_interspace = 0
        for key in self._garc_dict.keys():
            size = self._garc_dict[key].size
            self._garc_dict[key].coordinates = [None, None]
            self._garc_dict[key].coordinates[0] = (
                sum_interspace + start + ((end - start) * s / sum_length)
            )  # self.theta_list[s:s+self._garc_dict[key]["size"]+1]
            self._garc_dict[key].coordinates[1] = (
                sum_interspace + start + ((end - start) * (s + size) / sum_length)
            )
            s = s + size
            sum_interspace += self._garc_dict[key].interspace

        # self.figure = plt.figure(figsize=self.figsize)
        if self.fig_is_ext:
            self.ax = self.figure.add_axes(
                [0, 0, self.figsize[0], self.figsize[1]], polar=True
            )
        else:
            self.ax = self.figure.add_axes([0, 0, 1, 1], polar=True)
        self.ax.set_theta_zero_location("N")
        self.ax.set_theta_direction(-1)
        self.ax.set_ylim(0, 1000)
        self.ax.spines["polar"].set_visible(False)
        self.ax.xaxis.set_ticks([])
        self.ax.xaxis.set_ticklabels([])
        self.ax.yaxis.set_ticks([])
        self.ax.yaxis.set_ticklabels([])

        for i, key in enumerate(self._garc_dict.keys()):
            pos = self._garc_dict[key].coordinates[0]
            width = (
                self._garc_dict[key].coordinates[-1]
                - self._garc_dict[key].coordinates[0]
            )
            height = abs(
                self._garc_dict[key].raxis_range[1]
                - self._garc_dict[key].raxis_range[0]
            )
            bottom = self._garc_dict[key].raxis_range[0]
            facecolor = self._garc_dict[key].facecolor
            edgecolor = self._garc_dict[key].edgecolor
            linewidth = self._garc_dict[key].linewidth
            # print(key, pos, pos+width)
            self.ax.bar(
                [pos],
                [height],
                bottom=bottom,
                width=width,
                facecolor=facecolor,
                linewidth=linewidth,
                edgecolor=edgecolor,
                align="edge",
            )
            if self._garc_dict[key].label_visible == True:
                rot = (
                    self._garc_dict[key].coordinates[0]
                    + self._garc_dict[key].coordinates[1]
                ) / 2
                rot = rot * 360 / (2 * np.pi)
                if 90 < rot < 270:
                    rot = 180 - rot
                else:
                    rot = -1 * rot
                height = bottom + height / 2 + self._garc_dict[key].labelposition
                self.ax.text(
                    pos + width / 2,
                    height,
                    self._garc_dict[key].label,
                    rotation=rot,
                    ha="center",
                    va="center",
                    fontsize=self._garc_dict[key].labelsize,
                )

    def setspine(
        self,
        garc_id,
        raxis_range=(550, 600),
        facecolor="#30303000",
        edgecolor="#303030",
        linewidth=0.75,
    ):
        """Sets the spines in the sector corresponding to the arc of
        the Garc class object specified by garc_id.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID must be present in Gcircle object.garc_dict.
            raxis_range (tuple, optional):
                Radial axis range where the line plot is drawn.
                Defaults to (550, 600).
            facecolor (str, optional):
                Color for the spines area. Default is "#30303000".
            edgecolor (str, optional):
                Edge color of the spines boundary area. Default is "#303030".
            linewidth (float, optional):
                Edge line width of the spines boundary area. Default is 0.75.

        Returns:
            None
        """

        pos = self._garc_dict[garc_id].coordinates[0]
        width = (
            self._garc_dict[garc_id].coordinates[-1]
            - self._garc_dict[garc_id].coordinates[0]
        )
        height = abs(raxis_range[1] - raxis_range[0])
        bottom = raxis_range[0]
        self.ax.bar(
            [pos],
            [height],
            bottom=bottom,
            width=width,
            facecolor=facecolor,
            linewidth=linewidth,
            edgecolor=edgecolor,
            align="edge",
            zorder=0,
        )

    def lineplot(
        self,
        garc_id,
        data,
        positions=None,
        raxis_range=(550, 600),
        rlim=None,
        linestyle="solid",
        linecolor=None,
        linewidth=1.0,
        spine=False,
    ):
        """Plot a line in the sector corresponding to the arc of the Garc class
        object specified by garc_id.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID should be in Gcircle
                object.garc_dict.
            data (list or numpy.ndarray):
                Numerical data to be used for plot generation.
            positions (list or numpy.ndarray, optional):
                The x coordinates of the values in data on the Garc class object
                when the plot is drawn on rectangular coordinates. Each coordinate
                value should be in the range 0 to size of the Garc class object
                specified by garc_id. If positions are not given, proper coordinate
                values are generated according to the length of data. The default
                is None.
            raxis_range (tuple, optional):
                Radial axis range where line plot is drawn. The default is
                (550, 600).
            rlim (tuple, optional):
                The top and bottom r limits in data coordinates. If rlim value
                is not given, the maximum value and the minimum value in data
                will be set to top and bottom, respectively. The default is None.
            linestyle (str, optional):
                Line style. The default is "solid". Possible line styles are documented at
                https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html
            linecolor (str or tuple, optional):
                Color of the line plot. If linecolor value is not given, the color
                will be set according to the default color set of matplotlib. To
                specify the opacity for a line color, please use `(r,g,b,a)` or
                `#FFFFFF` format. The default is None.
            linewidth (float, optional):
                Edge line width. The default is 1.0.
            spine (bool, optional):
                If True, spines of the Garc object are shown on the arc section.
                The default is False.

        Returns:
            None
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size - 1
        positions_all = np.linspace(start, end, len(data), endpoint=True)
        if positions is None:
            positions = positions_all
        else:
            new_positions = []
            for p in positions:
                new_positions.append(start + ((end - start) * p / size))
            positions = new_positions

        if raxis_range is None:
            raxis_range = raxis_range[0]
        bottom = raxis_range[0]
        top = raxis_range[1]

        if linecolor is None:
            linecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)]
            self.color_cycle += 1

        if rlim is None:
            rlim = (min(data) - 0.05 * abs(min(data)), max(data) + 0.05 * abs(max(data)))

        min_value = rlim[0]
        max_value = rlim[1]
        new_data = []
        new_positions = []
        new_data_array = []
        new_positions_array = []
        for p, v in zip(positions, data):
            if v > rlim[1] or v < rlim[0]:
                new_data_array.append(new_data)
                new_positions_array.append(new_positions)
                new_data = []
                new_positions = []
            else:
                new_data.append(v)
                new_positions.append(p)
        new_data_array.append(new_data)
        new_positions_array.append(new_positions)
        for data, positions in zip(new_data_array, new_positions_array):
            if len(positions) > 0:
                data = np.array(data) - min_value
                data = bottom + np.array(data * ((top - bottom) / (max_value - min_value)))
                self.ax.plot(positions, data, color=linecolor, linewidth=linewidth, linestyle=linestyle)

        if spine == True:
            self.setspine(garc_id, raxis_range)

    def fillplot(
        self,
        garc_id,
        data,
        positions=None,
        raxis_range=(550, 600),
        rlim=None,
        base_value=None,
        facecolor=None,
        edgecolor="#303030",
        linewidth=0.0,
        spine=False,
    ):
        """
        Fill a specified area in the sector corresponding to the arc of the
        Garc class object specified by garc_id.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID should be in Gcircle object.garc_dict.
            data (list or numpy.ndarray):
                Numerical data used for plot generation.
            positions (list or numpy.ndarray, optional):
                The x coordinates of the values in data on the Garc class object
                when the plot is drawn on the rectangular coordinates. Each
                coordinate value should be in the range 0 to the size of the Garc class
                object specified by garc_id. If positions are not given, proper
                coordinate values are generated according to the length of data.
                Default is None.
            raxis_range (tuple of int, optional):
                Radial axis range where line plot is drawn. Default is (550, 600).
            rlim (tuple of int, optional):
                The top and bottom r limits in data coordinates. If not given,
                the maximum and minimum values in data will be set to top and bottom, respectively.
                Default is (min(data), max(data)).
            base_value (float, optional):
                Base line height in data coordinates. The area between the base
                line and the data line is filled by facecolor. Default is None.
            facecolor (str or tuple, optional):
                Color for filling. Default is None.
            edgecolor (str or tuple, optional):
                Edge color of the filled area. Default is "#303030".
            linewidth (float, optional):
                Edge line width. The default is 0.0.
            spine (bool, optional):
                If True, spines of the Garc object are shown on the arc section.
                Default is False.

        Returns:
            None:
                This function does not return any value.
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size - 1
        positions_all = np.linspace(start, end, len(data), endpoint=True)
        if positions is None:
            positions = positions_all
        else:
            new_positions = []
            for p in positions:
                new_positions.append(start + ((end - start) * p / size))
            positions = new_positions

        if raxis_range is None:
            raxis_range = raxis_range[0]
        bottom = raxis_range[0]
        top = raxis_range[1]

        if facecolor is None:
            facecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)]
            self.color_cycle += 1

        if rlim is None:
            rlim = (
                min(data) - 0.05 * abs(min(data)),
                max(data) + 0.05 * abs(max(data)),
            )

        min_value = rlim[0]
        max_value = rlim[1]
        if base_value is None:
            base_value = min_value
        new_data = []
        new_positions = []
        new_data_array = []
        new_positions_array = []
        for p, v in zip(positions, data):
            if v > rlim[1] or v < rlim[0]:
                new_data_array.append(new_data)
                new_positions_array.append(new_positions)
                new_data = []
                new_positions = []
            else:
                new_data.append(v)
                new_positions.append(p)
        new_data_array.append(new_data)
        new_positions_array.append(new_positions)
        for data, positions in zip(new_data_array, new_positions_array):
            if len(positions) > 0:
                base_value = base_value - min_value
                base_value = bottom + base_value * (
                    (top - bottom) / (max_value - min_value)
                )
                data = np.array(data) - min_value
                data = bottom + np.array(
                    data * ((top - bottom) / (max_value - min_value))
                )
                self.ax.fill_between(
                    positions,
                    data,
                    base_value,
                    facecolor=facecolor,
                    linewidth=linewidth,
                    edgecolor=edgecolor,
                )

        if spine == True:
            self.setspine(garc_id, raxis_range)

    def scatterplot(
        self,
        garc_id,
        data,
        positions=None,
        raxis_range=(550, 600),
        rlim=None,
        markershape="o",
        markersize=5,
        facecolor=None,
        edgecolor="#303030",
        linewidth=0.0,
        spine=False,
    ):
        """Draws a scatter plot on the sector corresponding to the arc of the Garc class object.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID should be in Gcircle object.garc_dict.
            data (list or numpy.ndarray):
                Numerical data used for plot generation.
            positions (list or numpy.ndarray, optional):
                The x coordinates of the values in data on the Garc class object
                when the plot is drawn on rectangular coordinates. Each coordinate
                value should be in the range 0 to the size of the Garc class
                object specified by garc_id. If positions are not provided,
                proper coordinate values are generated according to the length of data. Default is None.
            raxis_range (tuple, optional):
                Radial axis range where the line plot is drawn. Default is (550, 600).
            rlim (tuple, optional):
                The top and bottom r limits in data coordinates. Default is
                (min(data), max(data)) if not provided.
            markershape (str, optional):
                Marker shape. Default is "o".
                Possible markers are listed at
                https://matplotlib.org/stable/gallery/lines_bars_and_markers/marker_reference.html.
            markersize (float or list of float, optional):
                Size(s) of the marker(s). Default is 5.
            facecolor (str, tuple or list, optional):
                Face color(s) of the markers. If a list is provided, its length
                should match the length of data. Default is None.
            edgecolor (str or tuple, optional):
                Edge color of the markers. Default is "#303030".
            linewidth (float, optional):
                Edge line width of the markers. Default is 0.0.
            spine (bool, optional):
                If True, spines of the Garc object are shown in the arc section.
                Default is False.

        Returns:
            None
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size - 1
        positions_all = np.linspace(start, end, len(data), endpoint=True)
        if positions is None:
            positions = positions_all
        else:
            new_positions = []
            for p in positions:
                new_positions.append(start + ((end - start) * p / size))
            positions = new_positions

        if raxis_range is None:
            raxis_range = raxis_range[0]
        bottom = raxis_range[0]
        top = raxis_range[1]

        if facecolor is None:
            facecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)]
            self.color_cycle += 1

        if rlim is None:
            rlim = (min(data) - 0.05 * abs(min(data)), max(data) + 0.05 * abs(max(data)))

        min_value = rlim[0]
        max_value = rlim[1]
        new_data = []
        new_positions = []
        new_data_array = []
        new_positions_array = []
        for p, v in zip(positions, data):
            if v > rlim[1] or v < rlim[0]:
                new_data_array.append(new_data)
                new_positions_array.append(new_positions)
                new_data = []
                new_positions = []
            else:
                new_data.append(v)
                new_positions.append(p)

        new_data_array.append(new_data)
        new_positions_array.append(new_positions)
        for positions, data in zip(new_positions_array, new_data_array):
            if len(positions) > 0:
                data = np.array(data) - min_value
                data = bottom + np.array(data * ((top - bottom) / (max_value - min_value)))
                self.ax.scatter(positions, data, c=facecolor, s=markersize, linewidth=linewidth, edgecolor=edgecolor,
                                marker=markershape)

        if spine == True:
            self.setspine(garc_id, raxis_range)

    def barplot(
        self,
        garc_id,
        data,
        positions=None,
        width=None,
        raxis_range=(550, 600),
        rlim=None,
        base_value=None,
        facecolor=None,
        edgecolor="#303030",
        linewidth=0.0,
        spine=False,
    ):
        """Draws a bar plot within the specified arc of the Garc class object.

        This method visualizes numerical data as bars in a sector corresponding
        to the specified arc of a Garc class object identified by `garc_id`.

        Args:
            garc_id (str): The ID of the Garc class object. It should exist
                in the Gcircle object's garc_dict.
            data (list or numpy.ndarray): The numerical data used to generate
                the plot.
            positions (list or numpy.ndarray, optional): The x-coordinates of
                the values in `data` when plotted in rectangular coordinates.
                Each value should be within the range of the Garc class object's
                size identified by `garc_id`. If not provided, coordinates are
                automatically generated based on the length of `data`.
            width (float or list of float, optional): The width(s) of the bars.
                The default width is determined by `garc_object.size / len(data)`.
            raxis_range (tuple of int, optional): Range for the radial axis
                where the line plot is drawn, defaulting to (550, 600).
            rlim (tuple of int, optional): Specifies the top and bottom limits
                for the radial data coordinates. If not provided, it defaults
                to the minimum and maximum values of `data`.
            base_value (float, optional): Height of the baseline in data
                coordinates; the area between this line and the data line is filled with
                `facecolor`. Defaults to None.
            facecolor (str or tuple or list, optional): Specifies the face color(s)
                of the bars. If a list is provided, its length must match that of `data`.
                Default is None.
            edgecolor (str or tuple, optional): Color of the edges of the bars,
                defaulting to "#303030".
            linewidth (float, optional): Width of the edge lines of the bars;
                defaults to 0.0 (no lines).
            spine (bool, optional): If True, displays the spines of the Garc
                object on the arc section. Defaults to False.

        Returns:
            None: This method does not return any value.
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size
        positions_all = np.linspace(start, end, len(data), endpoint=False)
        if positions is None:
            positions = positions_all
        else:
            new_positions = []
            for p in positions:
                new_positions.append(start + ((end - start) * p / size))
            positions = new_positions

        if width is None:
            width = [positions[1] - positions[0]] * len(data)
        elif type(width) == float or type(width) == int:
            width = [(end - start) * width / size] * len(data)
        else:
            new_width = []
            for w in width:
                new_w = (end - start) * w / size
                new_width.append(new_w)
            width = new_width

        if raxis_range is None:
            raxis_range = raxis_range[0]
        bottom = raxis_range[0]
        top = raxis_range[1]

        if facecolor is None:
            facecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)]
            self.color_cycle += 1

        if rlim is None:
            if min(data) != max(data):
                rlim = (
                    min(data) - 0.05 * abs(min(data)),
                    max(data) + 0.05 * abs(max(data)),
                )
            else:
                rlim = (min(data), max(data))

        min_value = rlim[0] if rlim[0] is not None else min(data)
        max_value = rlim[1] if rlim[1] is not None else max(data)
        if base_value is None:
            base_value = min_value

        new_data = []
        new_positions = []
        new_width = []
        new_data_array = []
        new_positions_array = []
        new_width_array = []
        for p, v, w in zip(positions, data, width):
            if v > rlim[1] or v < rlim[0]:
                new_data_array.append(new_data)
                new_positions_array.append(new_positions)
                new_width_array.append(new_width)
                new_data = []
                new_width = []
                new_positions = []
            else:
                new_data.append(v)
                new_positions.append(p)
                new_width.append(w)

        new_data_array.append(new_data)
        new_positions_array.append(new_positions)
        new_width_array.append(new_width)
        for data, positions, width in zip(
            new_data_array, new_positions_array, new_width_array
        ):
            if len(positions) > 0:
                base_value = base_value - min_value
                if min_value != max_value:
                    base_value = bottom + base_value * (
                        (top - bottom) / (max_value - min_value)
                    )
                else:
                    base_value = raxis_range[0]

                data = np.array(data) - min_value
                if min_value != max_value:
                    data = np.array(data) * ((top - bottom) / (max_value - min_value))
                    data = np.array(data) - (base_value - raxis_range[0])
                else:
                    data = [raxis_range[1] - raxis_range[0]] * len(data)
                self.ax.bar(
                    positions,
                    data,
                    width=width,
                    bottom=base_value,
                    color=facecolor,
                    linewidth=linewidth,
                    edgecolor=edgecolor,
                    align="edge",
                )

        if spine == True:
            self.setspine(garc_id, raxis_range)

    def heatmap(
        self,
        garc_id,
        data,
        positions=None,
        width=None,
        raxis_range=(550, 600),
        cmap=None,
        vmin=None,
        vmax=None,
        edgecolor="#303030",
        linewidth=0.0,
        spine=False,
    ):
        """
        Visualize magnitudes of data values by color scale in the sector
        corresponding to the arc of the Garc class object specified by garc_id.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID should be in Gcircle object.garc_dict.
            data (list or numpy.ndarray):
                Numerical data to be used for plot generation.
            positions (list or numpy.ndarray, optional):
                The x coordinates of the values in data on the Garc class object
                when the plot is drawn on the rectangular coordinates. Each
                coordinate value should be in the range 0 to size of the Garc class
                object specified by garc_id. If positions are not given, proper
                coordinates values are generated according to the length of data.
                Defaults to None.
            width (float or list of float, optional):
                Width(s) of the bars. Defaults to `garc_object.size / len(data)`.
            raxis_range (tuple, optional):
                Radial axis range where heatmap is drawn. Default is (550, 600).
            cmap (str or matplotlib.colors.Colormap, optional):
                The mapping from data values to color space. Default is 'Reds'.
            vmin (float, optional):
                Minimum data threshold for color scale. Defaults to min(data).
            vmax (float, optional):
                Maximum data threshold for color scale. Defaults to max(data).
            edgecolor (str or tuple, optional):
                Edge color of the bars. Default is "#303030".
            linewidth (float, optional):
                Edge line width of the bars. Default is 0.0.
            spine (bool, optional):
                If True, spines of the Garc object are shown on the arc section.
                Default is False.

        Returns:
            None
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size
        positions_all = np.linspace(start, end, len(data), endpoint=False)
        if positions is None:
            positions = positions_all
        else:
            new_positions = []
            for p in positions:
                new_positions.append(start + ((end - start) * p / size))
            positions = new_positions

        if width is None:
            width = [positions[1] - positions[0]] * len(data)
        elif type(width) == float or type(width) == int:
            width = [(end - start) * width / size] * len(data)
        else:
            new_width = []
            for w in width:
                new_w = (end - start) * w / size
                new_width.append(new_w)
            width = new_width

        if raxis_range is None:
            raxis_range = raxis_range[0]
        bottom = raxis_range[0]
        top = raxis_range[1]
        height = top - bottom

        if cmap is None:
            cmap = Gcircle.cmaps[self.cmap_cycle % len(Gcircle.cmaps)]
            self.cmap_cycle += 1

        if vmax is None:
            max_value = max(data)
        else:
            max_value = vmax

        if vmin is None:
            min_value = min(data)
        else:
            min_value = vmin

        facecolors = []
        for d in data:
            facecolors.append(cmap(d / (max_value - min_value)))
        self.ax.bar(positions, height=[height] * len(positions), width=width, bottom=bottom, color=facecolors,
                    edgecolor=edgecolor, linewidth=linewidth, align="edge")

        if spine == True:
            self.setspine(garc_id, raxis_range)

    def featureplot(
        self,
        garc_id,
        feature_type=None,
        source=None,
        raxis_range=(550, 600),
        facecolor=None,
        edgecolor="#303030",
        linewidth=0.0,
        spine=False,
    ):
        """
        Visualize sequence features with bar plots in the sector corresponding
        to the arc of the Garc class object specified by garc_id.

        Args:
            garc_id (str):
                ID of the Garc class object. The ID should be in Gcircle object.garc_dict.
            feature_type (str, optional):
                Biological nature of the Bio.Seqfeature class objects.
                Accepts any value, but GenBank format requires registering a biological
                nature category for each sequence feature. If the value is "all",
                all features in source will be drawn in the sector of the Garc
                class object specified by garc_id. The default is 'all'.
            source (list of Bio.SeqFeature, optional):
                List of Bio.SeqFeature class objects. If not provided,
                record.features of the Garc class object specified by garc_id is used.
                The default is record.features of the Garc class object specified by garc_id.
            raxis_range (tuple of int, optional):
                Radial axis range where the feature plot is drawn.
                The default is (550, 600).
            facecolor (str or tuple, optional):
                Facecolor(s) of the bars. If a list is provided, its length
                should match the data length. The default is None.
            edgecolor (str or tuple, optional):
                Edge color of the bars. The default is "#303030".
            linewidth (float, optional):
                Edge line width of the bars. The default is 0.0.
            spine (bool, optional):
                If True, spines of the Garc object are shown on the arc section.
                The default is False.

        Returns:
            None
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size - 1

        if source is None:
            source = self.record.features

        if feature_type is None:
            feature_list = source
        else:
            feature_list = [feat for feat in source if feat.type == feature_type]

        positions = []
        widths = []
        for feat in feature_list:
            if feat.location.strand >= 0:
                s = int(feat.location.parts[0].start.position)
                e = int(feat.location.parts[-1].end.position)
                pos = start + ((end - start) * s / size)
                width = start + ((end - start) * e / size) - pos
                positions.append(pos)
                widths.append(width)
            else:
                s = int(feat.location.parts[-1].start.position)
                e = int(feat.location.parts[0].end.position)
                pos = start + ((end - start) * s / size)
                width = start + ((end - start) * e / size) - pos
                positions.append(pos)
                widths.append(width)

        bottom = raxis_range[0]
        top = raxis_range[1]

        if facecolor is None:
            facecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)]
            self.color_cycle += 1
        self.ax.bar(
            positions,
            [abs(top - bottom)] * len(positions),
            width=widths,
            bottom=bottom,
            color=facecolor,
            edgecolor=edgecolor,
            linewidth=linewidth,
            align="edge",
        )
        if spine == True:
            self.setspine(garc_id, raxis_range)

    def chord_plot(
        self, start_list, end_list, facecolor=None, edgecolor=None, linewidth=0.0
    ):
        """Visualize interrelationships between data.

        Args:
            start_list (tuple): Start data location of linked data. The tuple is composed of four parameters:
                - `arc_id` (str): The ID of the first Garc class object to be compared. The ID should be in `Gcircle` object's `garc_dict`.
                - `edge_position1` (int): The minimal x coordinate on the Garc class object when the plot is drawn on the rectangular coordinates.
                - `edge_position2` (int): The maximal x coordinate on the Garc class object when the plot is drawn on the rectangular coordinates.
                - `raxis_position` (int): The base height for the drawing chord.

            end_list (tuple): End data location of linked data. The tuple is composed of four parameters:
                - `arc_id` (str): The ID of the second Garc class object to be compared. the ID should be in `Gcircle` object's `garc_dict`.
                - `edge_position1` (int): The minimal x coordinate on the Garc class object when the plot is drawn on the rectangular coordinates.
                - `edge_position2` (int): The maximal x coordinate on the Garc class object when the plot is drawn on the rectangular coordinates.
                - `raxis_position` (int): The base height for the drawing chord.

            facecolor (str or tuple, optional):
                Facecolor of the link. The default is None.
            edgecolor (str or tuple, optional):
                Edge color of the link. The default is "#303030".
            linewidth (float, optional):
                Edge line width of the link. The default is 0.0.

        Returns:
            None
        """

        garc_id1 = start_list[0]
        garc_id2 = end_list[0]
        center = 0

        start1 = self._garc_dict[garc_id1].coordinates[0]
        end1 = self._garc_dict[garc_id1].coordinates[-1]
        size1 = self._garc_dict[garc_id1].size - 1
        sstart = start1 + ((end1 - start1) * start_list[1] / size1)
        send = start1 + ((end1 - start1) * start_list[2] / size1)
        stop = start_list[3]

        start2 = self._garc_dict[garc_id2].coordinates[0]
        end2 = self._garc_dict[garc_id2].coordinates[-1]
        size2 = self._garc_dict[garc_id2].size - 1
        ostart = start2 + ((end2 - start2) * end_list[1] / size2)
        oend = start2 + ((end2 - start2) * end_list[2] / size2)
        etop = end_list[3]

        if facecolor is None:
            facecolor = Gcircle.colors[self.color_cycle % len(Gcircle.colors)] + "80"
            self.color_cycle += 1

        z1 = stop - stop * math.cos(abs((send - sstart) * 0.5))
        z2 = etop - etop * math.cos(abs((oend - ostart) * 0.5))
        if sstart == ostart:
            pass
        else:
            Path = mpath.Path
            path_data = [(Path.MOVETO, (sstart, stop)),
                         (Path.CURVE3, (sstart, center)),
                         (Path.CURVE3, (oend, etop)),
                         (Path.CURVE3, ((ostart + oend) * 0.5, etop + z2)),
                         (Path.CURVE3, (ostart, etop)),
                         (Path.CURVE3, (ostart, center)),
                         (Path.CURVE3, (send, stop)),
                         (Path.CURVE3, ((sstart + send) * 0.5, stop + z1)),
                         (Path.CURVE3, (sstart, stop)),
                         ]
            codes, verts = list(zip(*path_data))
            path = mpath.Path(verts, codes)
            patch = mpatches.PathPatch(path, facecolor=facecolor, linewidth=linewidth, edgecolor=edgecolor, zorder=0)
            self.ax.add_patch(patch)

    def tickplot(
        self,
        garc_id,
        raxis_range=None,
        tickinterval=1000,
        tickpositions=None,
        ticklabels=None,
        tickwidth=1,
        tickcolor="#303030",
        ticklabelsize=10,
        ticklabelcolor="#303030",
        ticklabelmargin=10,
        tickdirection="outer",
        ticklabelorientation="vertical",
    ):
        """Plots ticks on the arc of the Garc class object.

        Args:
            garc_id (str):
                The ID of the Garc class object. The ID should be in
                Gcircle object.garc_dict.
            raxis_range (tuple of int, optional):
                Radial axis range where tick plot is drawn.
                If direction is "inner", the default is
                `(r0 - 0.5 * abs(r1 - r0), r0)`. If direction is
                "outer", the default is `(r1, r1 + 0.5 * abs(r1 - r0))`.
                `r0` and `r1` are defined as
                `Garc_object.raxis_range[0], Garc_object.raxis_range[1]`.
            tickinterval (int, optional):
                The interval between ticks. The default value is 1000.
                If `tickpositions` is provided, this value will be ignored.
            tickpositions (list of int, optional):
                Specific positions on the arc of the Garc class object
                where ticks should be placed. The values should be
                less than `Garc_object.size`.
            ticklabels (list of int or str, optional):
                Labels for the ticks on the arc of the Garc class object.
                The default value is the same as `tickpositions`.
            tickwidth (float, optional):
                The width of the ticks. The default value is 1.0.
            tickcolor (str or float, optional):
                The color of the ticks. The default value is "#303030".
            ticklabelsize (float, optional):
                The font size of the tick labels. The default value is 10.
            ticklabelcolor (str, optional):
                The color of the tick labels. The default value is "#303030".
            ticklabelmargin (float, optional):
                The margin of the tick labels. The default value is 10.
            tickdirection (str, optional):
                The direction of the ticks ("outer" or "inner").
                The default value is "outer".
            ticklabelorientation (str, optional):
                The orientation of the tick labels ("vertical" or "horizontal").
                The default value is "vertical".

        Returns:
            None
        """

        start = self._garc_dict[garc_id].coordinates[0]
        end = self._garc_dict[garc_id].coordinates[-1]
        size = self._garc_dict[garc_id].size + 1
        positions_all = np.linspace(start, end, size, endpoint=True)

        if raxis_range is None:
            r0, r1 = self._garc_dict[garc_id].raxis_range
            tickheight = 0.5 * abs(r1 - r0)
            if tickdirection == "outer":
                raxis_range = (r1, r1 + tickheight)
            elif tickdirection == "inner":
                raxis_range = (r0 - tickheight, r0)

        if tickpositions is None:
            tickpositions = [pos for pos in range(0, size, tickinterval)]

        if ticklabels is None:
            ticklabels = [None] * len(tickpositions)

        elif ticklabels == "None":
            ticklabels = tickpositions

        for pos, label in zip(tickpositions, ticklabels):
            self.ax.plot(
                [positions_all[pos], positions_all[pos]],
                raxis_range,
                linewidth=tickwidth,
                color=tickcolor,
            )
            if label is None:
                pass
            else:
                ticklabel_rot = self._get_label_rotation(
                    start + ((end - start) * (pos / size)), ticklabelorientation
                )
                if ticklabelorientation == "horizontal":
                    label_width = ticklabelsize * 2
                elif ticklabelorientation == "vertical":
                    label_width = ticklabelsize * len(str(label))

                if tickdirection == "outer":
                    y_pos = raxis_range[1] + (label_width + ticklabelmargin)
                elif tickdirection == "inner":
                    y_pos = raxis_range[0] - (label_width + ticklabelmargin)

                self.ax.text(
                    positions_all[pos],
                    y_pos,
                    str(label),
                    rotation=ticklabel_rot,
                    ha="center",
                    va="center",
                    fontsize=ticklabelsize,
                    color=ticklabelcolor,
                )

    def _get_label_rotation(self, position, orientation="horizontal"):
        """Calculates the rotation angle for a label based on its position and orientation.

        Args:
            position (float):
                The radian position of the label, which should be in the range
                (-2 * np.pi <= position <= 2 * np.pi).
            orientation (str):
                The orientation of the label, either "vertical" or "horizontal".
                Defaults to "horizontal".

        Returns:
            float:
                The calculated rotation angle for the label in degrees.
        """
        position_degree = position * (
            180 / np.pi
        )  # Convert radian to degree (-360 <= position_degree <= 360)
        if orientation == "horizontal":
            rotation = 0 - position_degree
            if -270 <= position_degree < -90 or 90 <= position_degree < 270:
                rotation += 180
        elif orientation == "vertical":
            rotation = 90 - position_degree
            if -180 <= position_degree < 0 or 180 <= position_degree < 360:
                rotation += 180
        return rotation

    def save(self, file_name="test", format="pdf", dpi=None):
        """Saves the figure object of the Gcircle class to a file.

        Args:
            file_name (str, optional):
                The name of the file to save the figure. Defaults to "test".
            format (str, optional):
                The file format for the saved figure. Defaults to "pdf".
            dpi (int, optional):
                The resolution of the saved figure in dots per inch. Defaults to None.

        Returns:
            None
        """

        self.figure.patch.set_alpha(0.0)
        if format == "pdf" and dpi is None:
            self.figure.savefig(file_name + ".pdf", bbox_inches="tight")
        else:
            if dpi is None:
                dpi = 600
            self.figure.savefig(file_name + "." + format, bbox_inches="tight", dpi=dpi)
        return self.figure


class Tarc(Garc):
    def __init__(
        self,
        arc_id=None,
        tree=None,
        format="newick",
        interspace=3,
        raxis_range=(900, 950),
        facecolor=None,
        edgecolor="#303030",
        linewidth=0,
        label=None,
        labelposition=0,
        labelsize=10,
        label_visible=False,
    ):
        """
        Initializes a Garc class object.

        Args:
            arc_id (str, optional):
                Unique identifier for the Garc class object. If not provided,
                an original unique ID is automatically generated. Default is None.
            tree (str):
                File name of phylogenetic tree.
            format (str):
                Format of the phylogenetic tree. Default is "newick".
            interspace (float, optional):
                Distance angle (degrees) to the adjacent arc section in clockwise
                sequence. The actual interspace size is determined by the ratio
                of size to the combined sum of the size and interspace values of
                the Garc class objects in the Gcircle class object. Default is 3.
            raxis_range (tuple, optional):
                Radial axis range where the line plot is drawn. Default is (900, 950).
            facecolor (str or tuple, optional):
                Color for filling. Default is None.
            edgecolor (str or tuple, optional):
                Edge color of the filled area. Default is "#303030".
            linewidth (float, optional):
                Edge line width. Default is 0.
            label (str, optional):
                Label of the arc section. Default is None.
            labelposition (int, optional):
                Relative label height from the center of the arc section. Default is 0.
            labelsize (int, optional):
                Font size of the label. Default is 10.
            label_visible (bool, optional):
                If True, the label of the Garc object is shown on the arc section.
                Default is False.

        Returns:
            None
        """

        self.tree = Phylo.read(tree, format)
        self.size = len(self.tree.get_terminals())
        self._tree_plotted = 0
        self._parental_gcircle = None
        if arc_id == None:
            self.arc_id = str(Garc._arcnum)
        else:
            self.arc_id = arc_id

        self.interspace = 2 * np.pi * (interspace / 360)
        self.raxis_range = raxis_range
        self.facecolor = facecolor
        self.edgecolor = edgecolor
        self.linewidth = linewidth

        if label is None:
            self.label = arc_id
        else:
            self.label = label

        self.label_visible = label_visible
        self.labelposition = labelposition
        self.labelsize = labelsize

        self._get_col_positions()
        self._get_row_positions()

        Garc._arcnum += 1

    def _get_col_positions(self):
        taxa = self.tree.get_terminals()
        depths = self.tree.depths()
        max_label_width = max(len(str(taxon)) for taxon in taxa)
        drawing_width = 100 - max_label_width - 1
        if max(depths.values()) == 0:
            depths = self.tree.depths(unit_branch_lengths=True)
        fudge_margin = math.log(len(taxa), 2)
        cols_per_branch_unit = (drawing_width - fudge_margin) / float(
            max(depths.values())
        )
        positions = {
            clade: blen * cols_per_branch_unit + 1.0 for clade, blen in depths.items()
        }
        self._col_positions = positions

    def _get_row_positions(self):
        taxa = self.tree.get_terminals()
        positions = {taxon: 2 * idx for idx, taxon in enumerate(taxa)}

        def calc_row(clade):
            for subclade in clade:
                if subclade not in positions:
                    calc_row(subclade)
            positions[clade] = (
                positions[clade.clades[0]] + positions[clade.clades[-1]]
            ) // 2

        calc_row(self.tree.root)
        self._row_positions = positions

        self.clade_dict = {}
        self.terminal_dict = {}
        self.nonterminal_dict = {}
        keys = list(self._row_positions.keys())
        keys.sort(key=lambda x: self._row_positions[x])
        for key in keys:
            if key in taxa:
                self.terminal_dict[key.name] = key
            else:
                self.nonterminal_dict[key.name] = key
            self.clade_dict[key.name] = key

    def _convert(self, thetalim, rlim):
        col_values = list(self._col_positions.values())
        row_values = list(self._row_positions.values())
        row_min, row_max = np.min(row_values), np.max(row_values)
        col_min, col_max = np.min(col_values), np.max(col_values)

        theta_positions = [
            thetalim[0] + (v * abs(thetalim[1] - thetalim[0]) / abs(row_min - row_max))
            for v in row_values
        ]
        if rlim[0] < rlim[1]:
            r_positions = [
                rlim[0] + (v * abs(rlim[1] - rlim[0]) / abs(col_min - col_max))
                for v in col_values
            ]
        else:
            r_positions = [
                rlim[0] - (v * abs(rlim[1] - rlim[0]) / abs(col_min - col_max))
                for v in col_values
            ]
        self._theta_dict = dict(zip(list(self._row_positions.keys()), theta_positions))
        self._r_dict = dict(zip(list(self._col_positions.keys()), r_positions))

    def _plot_tree(
        self,
        ax,
        thetalim=None,
        rlim=None,
        cladevisual_dict=None,
        highlight_dict=None,
        linewidth=None,
        linecolor=None,
    ):
        if linewidth is None:
            linewidth = 0.5

        if linecolor is None:
            linecolor = "k"

        if cladevisual_dict is None:
            cladevisual_dict = {}

        self._tree_rlim = rlim
        if rlim[0] < rlim[1]:
            self._tree_direction = "inner"
        else:
            self._tree_direction = "outer"

        self._convert(thetalim, rlim)
        s = []
        c = []
        ecs = []
        lws = []
        for clade in self._theta_dict:
            if clade.name not in cladevisual_dict:
                cladevisual_dict[clade.name] = {}
            cladevisual_dict[clade.name].setdefault("size", 0)
            cladevisual_dict[clade.name].setdefault("color", "k")
            cladevisual_dict[clade.name].setdefault("edgecolor", "k")
            cladevisual_dict[clade.name].setdefault("linewidth", 0.1)
            s.append(cladevisual_dict[clade.name]["size"])
            c.append(cladevisual_dict[clade.name]["color"])
            ecs.append(cladevisual_dict[clade.name]["edgecolor"])
            lws.append(cladevisual_dict[clade.name]["linewidth"])
        ax.scatter(
            self._theta_dict.values(),
            [self._r_dict[clade] for clade in self._theta_dict],
            s=s,
            c=c,
            edgecolors=ecs,
            linewidths=lws,
            zorder=1100,
        )
        for clade in self._theta_dict:
            subclades = clade.clades
            if len(subclades) > 0:
                sc_thetas = [self._theta_dict[sc] for sc in subclades]
                minsc_theta = min(sc_thetas)
                maxsc_theta = max(sc_thetas)
                thetas = np.linspace(minsc_theta, maxsc_theta, 100)
                rs = [self._r_dict[clade]] * len(thetas)
                ax.plot(thetas, rs, lw=linewidth, color=linecolor, zorder=0)
                for sc, sc_theta in zip(subclades, sc_thetas):
                    ax.plot(
                        [sc_theta, sc_theta],
                        [self._r_dict[sc], self._r_dict[clade]],
                        lw=linewidth,
                        color=linecolor,
                        zorder=0,
                    )

        if highlight_dict is not None:
            self.plot_highlight(ax, highlight_dict)
        self._tree_plotted = 1

    def _plot_highlight(self, ax, highlight_dict=None):
        if self._tree_plotted == 0:
            raise ValueError("Please run `plot_tree` before running `plot_highlight`")

        for clade_names in highlight_dict:
            highlight_dict[clade_names].setdefault("color", "#000000")
            highlight_dict[clade_names].setdefault("alpha", 0.25)
            highlight_dict[clade_names].setdefault("label", None)
            highlight_dict[clade_names].setdefault("y", None)
            highlight_dict[clade_names].setdefault("fontsize", self.labelsize)

            color = highlight_dict[clade_names]["color"]
            alpha = highlight_dict[clade_names]["alpha"]
            label = highlight_dict[clade_names]["label"]
            fontsize = highlight_dict[clade_names]["fontsize"]
            yloc = highlight_dict[clade_names]["y"]

            if type(clade_names) is tuple:
                ca = self.tree.common_ancestor(
                    [self.clade_dict[name] for name in clade_names]
                )
                clades = [self.clade_dict[name] for name in clade_names]
            else:
                ca = self.clade_dict[clade_names]
                clades = ca.get_terminals()

            c_thetas = [self._theta_dict[c] for c in clades]
            minc_theta = min(c_thetas)
            maxc_theta = max(c_thetas)

            if self._tree_direction == "inner":
                c_rs = [self._r_dict[c] for c in clades]
                maxc_r = max(c_rs)
                width = minc_theta - maxc_theta
                loc = (minc_theta + maxc_theta) / 2
                ax.bar(
                    [loc],
                    [self._tree_rlim[1] - self._r_dict[ca]],
                    bottom=self._r_dict[ca],
                    width=width,
                    color=color,
                    alpha=alpha,
                    linewidth=0,
                    zorder=1000 + maxc_r,
                )
                if highlight_dict[clade_names]["label"] is None:
                    pass
                else:
                    rot = loc * 360 / (2 * np.pi)
                    if 90 < rot < 270:
                        rot = 180 - rot
                    else:
                        rot = -1 * rot
                    if yloc is None:
                        yloc = (
                            self._tree_rlim[1]
                            - abs(self._tree_rlim[1] - self._tree_rlim[0]) * 0.1
                        )
                    ax.text(
                        loc,
                        yloc,
                        str(label),
                        rotation=rot,
                        ha="center",
                        va="center",
                        fontsize=fontsize,
                        zorder=1000 + maxc_r + 0.1,
                    )

            else:
                c_rs = [self._r_dict[c] for c in clades]
                minc_r = min(c_rs)
                width = minc_theta - maxc_theta
                loc = (minc_theta + maxc_theta) / 2
                ax.bar(
                    [loc],
                    [self._r_dict[ca] - self._tree_rlim[1]],
                    bottom=self._tree_rlim[1],
                    width=width,
                    color=color,
                    alpha=alpha,
                    linewidth=0,
                    zorder=1000 + (-1 * minc_r),
                )
                if highlight_dict[clade_names]["label"] is None:
                    pass
                else:
                    rot = loc * 360 / (2 * np.pi)
                    if 90 < rot < 270:
                        rot = 180 - rot
                    else:
                        rot = -1 * rot
                    if yloc is None:
                        yloc = (
                            self._tree_rlim[1]
                            + abs(self._tree_rlim[1] - self._tree_rlim[0]) * 0.1
                        )
                    ax.text(
                        loc,
                        yloc,
                        str(label),
                        rotation=rot,
                        ha="center",
                        va="center",
                        fontsize=fontsize,
                        zorder=1000 + (-1 * minc_r) + 0.1,
                    )


class Tcircle(Gcircle):
    """
    Tcircle class is the subclass of Gcircle. All methods implemented in the
    Gcircle class also can be used. Then, the two additional methods set_tarc,
    plot_tree and plot_highlight is provided in the Tcircle class.
    """

    def __init__(self, fig=None, figsize=None):
        """Initializes the circular mapping object.

        Args:
            fig (matplotlib.pyplot.figure, optional):
                Matplotlib Figure class object to be used.
            figsize (tuple, optional):
                Size of the figure for the circular map.
        """
        super().__init__(fig=fig, figsize=figsize)

    def __getattr__(self, name):
        """Retrieves attributes of the object.

        Args:
            name (str): The name of the attribute to retrieve.

        Returns:
            dict: The corresponding _garc_dict if name is 'tarc_dict'.
        """
        if name == "tarc_dict":
            return self._garc_dict

    def add_tarc(self, tarc):
        """Adds a new Tarc or Garc object to the tarc_dict.

        Args:
            tarc (Tarc or Garc):
                The Tarc or Garc object to be added.

        Returns:
            None
        """
        self._garc_dict[tarc.arc_id] = tarc

    def set_tarcs(self, start=0, end=360):
        """Visualizes the arc rectangles of Tarc objects in the .garc_dict.

        After execution, no new Tarc objects can be added to garc_dict,
        and a matplotlib.pyplot.figure will be created in the Tcircle object.

        Args:
            start (int, optional):
                Start angle of the circos plot, in the range of -360 to 360. Defaults to 0.
            end (int, optional):
                End angle of the circos plot, in the range of -360 to 360. Defaults to 360.

        Returns:
            None
        """

        sum_length = sum(list(map(lambda x: self._garc_dict[x]["size"], list(self._garc_dict.keys()))))
        sum_interspace = sum(list(map(lambda x: self._garc_dict[x]["interspace"], list(self._garc_dict.keys()))))
        start = 2 * np.pi * start / 360
        end = (2 * np.pi * end / 360) - sum_interspace

        s = 0
        sum_interspace = 0
        for key in self._garc_dict.keys():
            size = self._garc_dict[key].size
            self._garc_dict[key].coordinates = [None, None]
            self._garc_dict[key].coordinates[0] = sum_interspace + start + ((end - start) * s / sum_length)
            self._garc_dict[key].coordinates[1] = sum_interspace + start + ((end - start) * (s + size) / sum_length)
            s = s + size
            sum_interspace += self._garc_dict[key].interspace

        if self.fig_is_ext:
            self.ax = self.figure.add_axes([0, 0, self.figsize[0], self.figsize[1]], polar=True)
        else:
            self.ax = self.figure.add_axes([0, 0, 1, 1], polar=True)
        self.ax.set_theta_zero_location("N")
        self.ax.set_theta_direction(-1)
        self.ax.set_ylim(0, 1000)
        self.ax.spines['polar'].set_visible(False)
        self.ax.xaxis.set_ticks([])
        self.ax.xaxis.set_ticklabels([])
        self.ax.yaxis.set_ticks([])
        self.ax.yaxis.set_ticklabels([])

        for i, key in enumerate(self._garc_dict.keys()):
            pos = self._garc_dict[key].coordinates[0]
            width = self._garc_dict[key].coordinates[-1] - self._garc_dict[key].coordinates[0]
            height = abs(self._garc_dict[key].raxis_range[1] - self._garc_dict[key].raxis_range[0])
            bottom = self._garc_dict[key].raxis_range[0]
            facecolor = self._garc_dict[key].facecolor
            edgecolor = self._garc_dict[key].edgecolor
            linewidth = self._garc_dict[key].linewidth
            if facecolor is None:
                facecolor = (0, 0, 0, 0)

            if facecolor == (0, 0, 0, 0) and linewidth == 0:
                pass
            else:
                self.ax.bar([pos], [height], bottom=bottom, width=width, facecolor=facecolor, linewidth=linewidth,
                            edgecolor=edgecolor, align="edge")

            if self._garc_dict[key].label_visible == True:
                rot = (self._garc_dict[key].coordinates[0] + self._garc_dict[key].coordinates[1]) / 2
                rot = rot * 360 / (2 * np.pi)
                if 90 < rot < 270:
                    rot = 180 - rot
                else:
                    rot = -1 * rot
                height = bottom + height / 2 + self._garc_dict[key].labelposition
                self.ax.text(pos + width / 2, height, self._garc_dict[key].label, rotation=rot, ha="center",
                             va="center", fontsize=self._garc_dict[key].labelsize)

    def plot_tree(
        self,
        tarc_id,
        rlim=(0, 700),
        cladevisual_dict=None,
        highlight_dict=None,
        linecolor="#303030",
        linewidth=0.5,
    ):
        """Draw a circular phylogenetic tree.

        Args:
            tarc_id (str):
                The ID of the Tarc class object. The ID should be in
                Tcircle object.tarc_dict.
            rlim (tuple of int, optional):
                The top and bottom radial limits in data coordinates.
                Defaults to (0, 700).
            cladevisual_dict (dict, optional):
                A dictionary containing clade visualization parameters,
                structured as pairs of clade name and a sub-dictionary with
                the following keys:

                - `size` (float): Size of the dot. Defaults to 5.
                - `color` (str or float): Face color of the dot.
                  Defaults to "#303030".
                - `edgecolor` (str or float): Edge line color of the dot.
                  Defaults to "#303030".
                - `linewidth` (float): Edge line width of the dot.
                  Defaults to 0.5.

            highlight_dict (dict, optional):
                A dictionary containing clade highlight parameters,
                which can also use tuples of terminal clade names.
                The structure includes:

                - `color` (str): Color for highlighting clades.
                  Defaults to "#000000".
                - `alpha` (float): Transparency level for highlights.
                  Defaults to 0.25.
                - `label` (str): Label for the highlight. Defaults to None.
                - `fontsize` (float): Font size of the label.
                  Defaults to 10.
                - `y` (float): Y location of the text.
                  Defaults to the bottom edge of the highlight.

            linecolor (str or tuple, optional):
                Color of the tree line. Defaults to "#303030".
            linewidth (float):
                Line width of the tree. Defaults to 0.5.

        Returns:
            None
        """

        start = self._garc_dict[tarc_id].coordinates[0]
        end = self._garc_dict[tarc_id].coordinates[-1]
        positions = np.linspace(
            start, end, self._garc_dict[tarc_id].size, endpoint=False
        )
        positions = positions + abs(positions[1] - positions[0]) * 0.5
        start, end = positions[0], positions[-1]
        self._garc_dict[tarc_id]._plot_tree(
            self.ax,
            thetalim=(start, end),
            rlim=rlim,
            cladevisual_dict=cladevisual_dict,
            highlight_dict=highlight_dict,
            linecolor=linecolor,
            linewidth=linewidth,
        )

    def plot_highlight(self, tarc_id, highlight_dict=None):
        """Add highlight for a specific clade under the given internal clade.

        Args:
            tarc_id (str):
                ID of the Tarc class object. The ID should be in Tcircle object's
                tarc_dict.
            highlight_dict (dict, optional):
                A dictionary composed of pairs of internal clade name and a sub-dict.
                Instead of clade name, tuples of terminal clade names can also be used.
                A sub-dict is composed of the following key-value pairs:

                - color (str):
                    Color of the highlight for clades. The default is "#000000".
                - alpha (float):
                    Alpha of the highlight for clades. The default is 0.25.
                - label (str, optional):
                    Label for the highlight. The default is None.
                - fontsize (float):
                    Font size of the label. The default is 10.
                - y (float):
                    Y location of the text. The default is at the bottom edge of the highlight.

        Returns:
            None
        """

        self._garc_dict[tarc_id]._plot_highlight(self.ax, highlight_dict=highlight_dict)


def table_hg38_chromosome_length():
    return pd.read_csv(os.path.join(DATAPATH, 'hg38_chromosome_length.csv'))

def table_hg38_cytoband():
    return pd.read_csv(os.path.join(DATAPATH, 'hg38_cytoBand.csv'))

def table_mm10_chromosome_length():
    return pd.read_csv(os.path.join(DATAPATH, 'mm10_chromosome_length.csv'))

def table_mm10_cytoband():
    return pd.read_csv(os.path.join(DATAPATH, 'mm10_cytoBand.csv'))


if __name__ == "__main__":
    pass
    """
    tree = next(Phylo.parse("../tutorial/circos/hmptree.xml", "phyloxml")) 
    col_positions = get_col_positions(tree) 
    row_positions = get_row_positions(tree)

    import numpy as np 
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(5,5)) 
    ax  = fig.add_axes([0, 0, 1, 1], polar=True)
    theta_dict, r_dict = convert(tree, col_positions, row_positions, (0, np.pi), (0,10))

    #ax.scatter(theta_dict.values(), [r_dict[clade] for clade in theta_dict], s=1, color="k") 

    plot_lines(ax, theta_dict, r_dict) 

    ax.set_ylim([0,10])
    fig.savefig("test.pdf", bbox_inches="tight")
    """




from bioat.lib.libjgi import JGIOperator
from bioat.logger import LoggerManager

lm = LoggerManager(mod_name="bioat.metatools")


class MetaTools:
    """Metagenome toolbox."""

    lm.set_names(cls_name="MetaTools")

    def JGI_query(
        self,
        query_info: str | None = None,
        xml: str | None = None,
        log_fails: str | None = None,
        nretry: int = 4,
        timeout: int = 60,
        regex: str | None = None,
        all_get: bool = False,
        overwrite_conf: bool = False,
        filter_files: bool = False,
        proxy_pool: str | None = None,
        just_query_xml: bool = False,
        syntax_help: bool = False,
        usage: bool = False,
        log_level: str = "INFO",
    ):
        """JGI_query: Tool for downloading files from the JGI-IMG database.

        This function lists and retrieves files from JGI using the curl API and
        returns a list of all files available for download for a given query organism.

        The source code is adapted from https://github.com/glarue/jgi-query.

        Args:
            query_info (str | None):
                Organism name formatted per JGI's abbreviation.
                Example: 'Nematostella vectensis' is abbreviated by JGI as 'Nemve1'.
                The correct abbreviation can be found by searching for the organism on JGI;
                the name used in the URL of the 'Info' page for that organism is
                the correct abbreviation. The full URL may also be used for this argument.
            xml (str | None):
                Specify a local XML file for the query instead of retrieving
                a new copy from JGI.
            log_fails (str | None):
                Log file containing URLs to retry downloading from in case of failure.
            nretry (int):
                Number of times to retry downloading files with errors.
                Use 0 to skip such files.
            timeout (int):
                Timeout (in seconds) for downloading. Set to -1 to disable.
            regex (str | None):
                Regex pattern to use for auto-selecting and downloading files
                without interaction.
            all_get (bool):
                If True, auto-select and download all files for the query
                without interaction.
            overwrite_conf (bool):
                If True, initiate configuration dialog to overwrite
                existing user/password configuration.
            filter_files (bool):
                Under development. Filter organism results by config categories
                instead of reporting all files listed by JGI for the query.
            proxy_pool (str | None):
                URL for the proxy pool, e.g., http://abc.com:port. See
                https://github.com/hermanzhaozzzz/proxy_pool.
            just_query_xml (bool):
                Set True if you just want to save the XML file.
            syntax_help (bool):
                If True, provide syntax help in doc mode.
            usage (bool):
                If True, print verbose usage information and exit.
            log_level (str):
                Set the logging level. Options include 'CRITICAL', 'ERROR',
                'WARNING', 'INFO', 'DEBUG', 'NOTSET'.
        """
        # Load or create JGI account info and
        # auto check if user info needs to be overwritten.

        operator = JGIOperator(
            query_info=query_info,
            xml=xml,
            log_fails=log_fails,
            nretry=nretry,
            timeout=timeout,
            regex=regex,
            all_get=all_get,
            overwrite_conf=overwrite_conf,
            filter_files=filter_files,
            proxy_pool=proxy_pool,
            just_query_xml=just_query_xml,
            syntax_help=syntax_help,
            usage=usage,
            log_level=log_level,
        )
        lm.set_names(func_name="JGI_query")
        lm.set_log_level(log_level)

        lm.logger.debug("run query")
        operator.query()
        lm.logger.debug("parse xml to json")
        operator.parse_xml()
        lm.logger.debug(
            "start to download; calculate and display total size of selected data"
        )
        operator.download()
